#!/usr/bin/env python
# Copyright 2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Parsers for basis set in the NWChem format
'''

__all__ = ['parse', 'load', 'convert_ecp_to_nwchem']

import numpy
import numpy as np
import re
from pyscf.data.elements import _std_symbol
from pyscf.lib.exceptions import BasisNotFoundError
from pyscf import __config__

DISABLE_EVAL = getattr(__config__, 'DISABLE_EVAL', False)

MAXL = 15
SPDF = 'SPDFGHIKLMNORTU'
MAPSPDF = {key: l for l, key in enumerate(SPDF)}

ECP_DELIMITER = re.compile('\n *ECP *\n')

def parse(string, symb=None):
    '''Parse ECP text which is in NWChem format. Return an internal
    basis format which can be assigned to attribute :attr:`Mole.ecp`
    Empty lines, or the lines started with #, or the lines of "BASIS SET" and
    "END" will be ignored are ignored.
    '''

    if symb is not None:
        symb = _std_symbol(symb)
        raw_data = string.splitlines()
        for i, dat in enumerate(raw_data):
            dat0 = dat.split(None, 1)
            if dat0 and dat0[0] == symb:
                break
        if i+1 == len(raw_data):
            raise BasisNotFoundError('ECP not found for  %s' % symb)
        seg = []
        for dat in raw_data[i:]:
            dat = dat.strip()
            if dat: # remove empty lines
                if ((dat[0].isalpha() and dat.split(None, 1)[0].upper() != symb.upper())):
                    break
                else:
                    seg.append(dat)
    else:
        seg = string.splitlines()

    ecptxt = []
    for dat in seg:
        dat = dat.split('#')[0].strip()
        dat_upper = dat.upper()
        if (dat and not dat_upper.startswith('END') and not dat_upper.startswith('ECP')):
            ecptxt.append(dat)
    return _parse_ecp(ecptxt)

def _parse_ecp(raw_ecp):
    ecp_add = []
    nelec = None
    for line in raw_ecp:
        dat = line.strip()
        if not dat or dat.startswith('#'): # comment line
            continue
        elif dat[0].isalpha():
            keys = dat.split()
            if len(keys) == 1:
                key = keys[0].upper()
            else:
                key = keys[1].upper()
            if key == 'NELEC':
                nelec = int(dat.split()[2])
                continue
            elif key == 'UL':
                ecp_add.append([-1])
            elif key in MAPSPDF:
                ecp_add.append([MAPSPDF[key]])
            else:
                raise BasisNotFoundError('Not basis data')
            # up to r^6
            by_ang = [[] for i in range(7)]
            ecp_add[-1].append(by_ang)
        else:
            line = dat.replace('D','e').split()
            l = int(line[0])
            try:
                coef = [float(x) for x in line[1:]]
            except ValueError:
                if DISABLE_EVAL:
                    raise ValueError('Failed to parse ecp %s' % line)
                else:
                    coef = list(eval(','.join(line[1:])))
            except Exception:
                raise BasisNotFoundError('Not basis data')
            if any(x != 0 for x in coef[1:]):
                by_ang[l].append(coef)

    if nelec is None:
        return []

    if not ecp_add:
        raise BasisNotFoundError(f'ECP data not found in "{raw_ecp}"')

    ecp_sorted = sorted(ecp_add, key=lambda shell: shell[0])
    return [nelec, ecp_sorted]

def load(basisfile, symb):
    '''Load ECP for atom of symb from file'''
    return _parse_ecp(_search_ecp(basisfile, symb))

def _search_ecp(basisfile, symb):
    symb = _std_symbol(symb)
    with open(basisfile, 'r') as fin:
        fdata = re.split(ECP_DELIMITER, fin.read())
    if len(fdata) <= 1:
        return []

    fdata = fdata[1].splitlines()
    for i, dat in enumerate(fdata):
        dat0 = dat.split(None, 1)
        if dat0 and dat0[0] == symb:
            break
    seg = []
    for dat in fdata[i:]:
        dat = dat.strip()
        if dat:  # remove empty lines
            if ((dat[0].isalpha() and dat.split(None, 1)[0].upper() != symb.upper())):
                return seg
            else:
                seg.append(dat)
    return []

def convert_ecp_to_nwchem(symb, ecp):
    '''Convert the internal ecp format to NWChem format string'''
    symb = _std_symbol(symb)
    res = ['%-2s nelec %d' % (symb, ecp[0])]

    for ecp_block in ecp[1]:
        l = ecp_block[0]
        if l == -1:
            res.append('%-2s ul' % symb)
        else:
            res.append('%-2s %s' % (symb, SPDF[l].lower()))
        for r_order, dat in enumerate(ecp_block[1]):
            for e,c in dat:
                res.append('%d    %15.9f  %15.9f' % (r_order, e, c))
    return '\n'.join(res)



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Generate SCF response functions
'''

import warnings
import numpy
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import hf, rohf, uhf, ghf, dhf

def _gen_rhf_response(mf, mo_coeff=None, mo_occ=None,
                      singlet=None, hermi=0, max_memory=None, with_nlc=True):
    '''Generate a function to compute the product of RHF response function and
    RHF density matrices.

    Kwargs:
        singlet (None or boolean) : If singlet is None, response function for
            orbital hessian or CPHF will be generated. If singlet is boolean,
            it is used in TDDFT response kernel.
        with_nlc (boolean) : NLC contribution is typically very small. This flag
        allows to skip NLC contribution.
    '''
    assert isinstance(mf, hf.RHF) and not isinstance(mf, (uhf.UHF, rohf.ROHF))

    if mo_coeff is None: mo_coeff = mf.mo_coeff
    if mo_occ is None: mo_occ = mf.mo_occ
    mol = mf.mol
    if isinstance(mf, hf.KohnShamDFT):
        ni = mf._numint
        ni.libxc.test_deriv_order(mf.xc, 2, raise_error=True)
        omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, mol.spin)
        hybrid = ni.libxc.is_hybrid_xc(mf.xc)

        if singlet is None: # for ground state orbital hessian
            spin = 0
        else:
            spin = 1
        rho0, vxc, fxc = ni.cache_xc_kernel(mol, mf.grids, mf.xc,
                                            mo_coeff, mo_occ, spin)
        dm0 = None

        if max_memory is None:
            mem_now = lib.current_memory()[0]
            max_memory = max(2000, mf.max_memory*.8-mem_now)

        if singlet is None:
            # Without specify singlet, used in ground state orbital hessian
            def vind(dm1):
                # The singlet hessian
                if hermi == 2:
                    v1 = numpy.zeros_like(dm1)
                else:
                    v1 = ni.nr_rks_fxc(mol, mf.grids, mf.xc, dm0, dm1, 0, hermi,
                                       rho0, vxc, fxc, max_memory=max_memory)
                    if with_nlc and mf.do_nlc():
                        from pyscf.hessian.rks import get_vnlc_resp # Cannot import at top due to circular dependency
                        v1 += get_vnlc_resp(mf, mol, mo_coeff, mo_occ, dm1, max_memory)
                if hybrid:
                    if omega == 0:
                        vj, vk = mf.get_jk(mol, dm1, hermi)
                        vk *= hyb
                    elif alpha == 0: # LR=0, only SR exchange
                        vj = mf.get_j(mol, dm1, hermi)
                        vk = mf.get_k(mol, dm1, hermi, omega=-omega)
                        vk *= hyb
                    elif hyb == 0: # SR=0, only LR exchange
                        vj = mf.get_j(mol, dm1, hermi)
                        vk = mf.get_k(mol, dm1, hermi, omega=omega)
                        vk *= alpha
                    else: # SR and LR exchange with different ratios
                        vj, vk = mf.get_jk(mol, dm1, hermi)
                        vk *= hyb
                        vk += mf.get_k(mol, dm1, hermi, omega=omega) * (alpha-hyb)
                    if hermi != 2:
                        v1 += vj - .5 * vk
                    else:
                        v1 += -.5 * vk
                elif hermi != 2:
                    v1 += mf.get_j(mol, dm1, hermi=hermi)
                return v1

        elif singlet:
            fxc *= .5
            def vind(dm1):
                if hermi == 2:
                    v1 = numpy.zeros_like(dm1)
                else:
                    # nr_rks_fxc_st requires alpha of dm1, dm1*.5 should be scaled
                    v1 = ni.nr_rks_fxc_st(mol, mf.grids, mf.xc, dm0, dm1, hermi, True,
                                          rho0, vxc, fxc, max_memory=max_memory)
                    if with_nlc and mf.do_nlc():
                        from pyscf.hessian.rks import get_vnlc_resp # Cannot import at top due to circular dependency
                        v1 += get_vnlc_resp(mf, mol, mo_coeff, mo_occ, dm1, max_memory)
                if hybrid:
                    if omega == 0:
                        vj, vk = mf.get_jk(mol, dm1, hermi)
                        vk *= hyb
                    elif alpha == 0: # LR=0, only SR exchange
                        vj = mf.get_j(mol, dm1, hermi)
                        vk = mf.get_k(mol, dm1, hermi, omega=-omega)
                        vk *= hyb
                    elif hyb == 0: # SR=0, only LR exchange
                        vj = mf.get_j(mol, dm1, hermi)
                        vk = mf.get_k(mol, dm1, hermi, omega=omega)
                        vk *= alpha
                    else: # SR and LR exchange with different ratios
                        vj, vk = mf.get_jk(mol, dm1, hermi)
                        vk *= hyb
                        vk += mf.get_k(mol, dm1, hermi, omega=omega) * (alpha-hyb)
                    if hermi != 2:
                        v1 += vj - .5 * vk
                    else:
                        v1 += -.5 * vk
                elif hermi != 2:
                    v1 += mf.get_j(mol, dm1, hermi=hermi)
                return v1
        else:  # triplet
            fxc *= .5
            def vind(dm1):
                if hermi == 2:
                    v1 = numpy.zeros_like(dm1)
                else:
                    # nr_rks_fxc_st requires alpha of dm1, dm1*.5 should be scaled
                    v1 = ni.nr_rks_fxc_st(mol, mf.grids, mf.xc, dm0, dm1, hermi, False,
                                          rho0, vxc, fxc, max_memory=max_memory)
                    if with_nlc and mf.do_nlc():
                        pass # fxc = 0, do nothing
                if hybrid:
                    if omega == 0:
                        vk = mf.get_k(mol, dm1, hermi) * hyb
                    elif alpha == 0: # LR=0, only SR exchange
                        vk = mf.get_k(mol, dm1, hermi, omega=-omega) * hyb
                    elif hyb == 0: # SR=0, only LR exchange
                        vk = mf.get_k(mol, dm1, hermi, omega=omega) * alpha
                    else: # SR and LR exchange with different ratios
                        vk = mf.get_k(mol, dm1, hermi) * hyb
                        vk += mf.get_k(mol, dm1, hermi, omega=omega) * (alpha-hyb)
                    v1 += -.5 * vk
                return v1

    else:  # HF
        if (singlet is None or singlet) and hermi != 2:
            def vind(dm1):
                vj, vk = mf.get_jk(mol, dm1, hermi=hermi)
                return vj - .5 * vk
        else:
            def vind(dm1):
                return -.5 * mf.get_k(mol, dm1, hermi=hermi)

    return vind


def _gen_uhf_response(mf, mo_coeff=None, mo_occ=None,
                      with_j=True, hermi=0, max_memory=None, with_nlc=True):
    '''Generate a function to compute the product of UHF response function and
    UHF density matrices.
    '''
    assert isinstance(mf, (uhf.UHF, rohf.ROHF))
    if mo_coeff is None: mo_coeff = mf.mo_coeff
    if mo_occ is None: mo_occ = mf.mo_occ
    mol = mf.mol
    if isinstance(mf, hf.KohnShamDFT):
        ni = mf._numint
        ni.libxc.test_deriv_order(mf.xc, 2, raise_error=True)
        omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, mol.spin)
        hybrid = ni.libxc.is_hybrid_xc(mf.xc)

        rho0, vxc, fxc = ni.cache_xc_kernel(mol, mf.grids, mf.xc,
                                            mo_coeff, mo_occ, 1)
        dm0 = None

        if max_memory is None:
            mem_now = lib.current_memory()[0]
            max_memory = max(2000, mf.max_memory*.8-mem_now)

        def vind(dm1):
            if hermi == 2:
                v1 = numpy.zeros_like(dm1)
            else:
                v1 = ni.nr_uks_fxc(mol, mf.grids, mf.xc, dm0, dm1, 0, hermi,
                                   rho0, vxc, fxc, max_memory=max_memory)
                if with_nlc and mf.do_nlc():
                    from pyscf.hessian.rks import get_vnlc_resp # Cannot import at top due to circular dependency
                    v1 += get_vnlc_resp(mf, mol, mo_coeff, mo_occ, dm1[0] + dm1[1], max_memory)
            if not hybrid:
                if with_j:
                    vj = mf.get_j(mol, dm1, hermi=hermi)
                    v1 += vj[0] + vj[1]
            else:
                if omega == 0:
                    vj, vk = mf.get_jk(mol, dm1, hermi, with_j=with_j)
                    vk *= hyb
                elif alpha == 0: # LR=0, only SR exchange
                    if with_j:
                        vj = mf.get_j(mol, dm1, hermi)
                    vk = mf.get_k(mol, dm1, hermi, omega=-omega)
                    vk *= hyb
                elif hyb == 0: # SR=0, only LR exchange
                    if with_j:
                        vj = mf.get_j(mol, dm1, hermi)
                    vk = mf.get_k(mol, dm1, hermi, omega=omega)
                    vk *= alpha
                else: # SR and LR exchange with different ratios
                    vj, vk = mf.get_jk(mol, dm1, hermi, with_j=with_j)
                    vk *= hyb
                    vk += mf.get_k(mol, dm1, hermi, omega=omega) * (alpha-hyb)
                if with_j:
                    v1 += vj[0] + vj[1] - vk
                else:
                    v1 -= vk
            return v1

    elif with_j:
        def vind(dm1):
            vj, vk = mf.get_jk(mol, dm1, hermi=hermi)
            v1 = vj[0] + vj[1] - vk
            return v1

    else:
        def vind(dm1):
            return -mf.get_k(mol, dm1, hermi=hermi)

    return vind


def _gen_ghf_response(mf, mo_coeff=None, mo_occ=None,
                      with_j=True, hermi=0, max_memory=None, with_nlc=True):
    '''Generate a function to compute the product of GHF response function and
    GHF density matrices.
    '''
    if mo_coeff is None: mo_coeff = mf.mo_coeff
    if mo_occ is None: mo_occ = mf.mo_occ
    mol = mf.mol
    if isinstance(mf, hf.KohnShamDFT):
        from pyscf.dft import numint2c, r_numint
        ni = mf._numint
        assert isinstance(ni, (numint2c.NumInt2C, r_numint.RNumInt))
        ni.libxc.test_deriv_order(mf.xc, 2, raise_error=True)
        omega, alpha, hyb = ni.rsh_and_hybrid_coeff(mf.xc, mol.spin)
        hybrid = ni.libxc.is_hybrid_xc(mf.xc)

        rho0, vxc, fxc = ni.cache_xc_kernel(mol, mf.grids, mf.xc, mo_coeff, mo_occ, 1)
        dm0 = None

        if max_memory is None:
            mem_now = lib.current_memory()[0]
            max_memory = max(2000, mf.max_memory*.8-mem_now)

        def vind(dm1):
            if hermi == 2:
                v1 = numpy.zeros_like(dm1)
            else:
                v1 = ni.get_fxc(mol, mf.grids, mf.xc, dm0, dm1, 0, 0, hermi,
                                rho0, vxc, fxc, max_memory=max_memory)
                if with_nlc and mf.do_nlc():
                    from pyscf.hessian.rks import get_vnlc_resp
                    nao = mo_coeff.shape[0] // 2
                    dm1_sf = dm1[...,:nao,:nao] + dm1[...,nao:,nao:]
                    mo_uks = [mo_coeff[:nao], mo_coeff[nao:]]
                    mo_occ_uks = [mo_occ, mo_occ]
                    # The get_vnlc_resp function uses mf._numint and explicitly
                    # calls the NumInt functions for mf._numint
                    ni1c = mf._numint._to_numint1c()
                    with lib.temporary_env(mf, _numint=ni1c):
                        vxc_nlc = get_vnlc_resp(mf, mol, mo_uks, mo_occ_uks,
                                                dm1_sf.real, max_memory)
                        v1[...,:nao,:nao] += vxc_nlc
                        v1[...,nao:,nao:] += vxc_nlc
            if not hybrid:
                if with_j:
                    vj = mf.get_j(mol, dm1, hermi=hermi)
                    v1 += vj
            else:
                if omega == 0:
                    vj, vk = mf.get_jk(mol, dm1, hermi, with_j=with_j)
                    vk *= hyb
                elif alpha == 0: # LR=0, only SR exchange
                    if with_j:
                        vj = mf.get_j(mol, dm1, hermi)
                    vk = mf.get_k(mol, dm1, hermi, omega=-omega)
                    vk *= hyb
                elif hyb == 0: # SR=0, only LR exchange
                    if with_j:
                        vj = mf.get_j(mol, dm1, hermi)
                    vk = mf.get_k(mol, dm1, hermi, omega=omega)
                    vk *= alpha
                else: # SR and LR exchange with different ratios
                    vj, vk = mf.get_jk(mol, dm1, hermi, with_j=with_j)
                    vk *= hyb
                    vk += mf.get_k(mol, dm1, hermi, omega=omega) * (alpha-hyb)
                if with_j:
                    v1 += vj - vk
                else:
                    v1 -= vk
            return v1

    elif with_j:
        def vind(dm1):
            vj, vk = mf.get_jk(mol, dm1, hermi=hermi)
            return vj - vk

    else:
        def vind(dm1):
            return -mf.get_k(mol, dm1, hermi=hermi)

    return vind


def _gen_dhf_response(mf, mo_coeff=None, mo_occ=None,
                      with_j=True, hermi=0, max_memory=None, with_nlc=True):
    '''Generate a function to compute the product of DHF response function and
    DHF density matrices.
    '''
    return _gen_ghf_response(mf, mo_coeff, mo_occ, with_j, hermi, max_memory)


hf.RHF.gen_response = _gen_rhf_response
uhf.UHF.gen_response = _gen_uhf_response
ghf.GHF.gen_response = _gen_ghf_response
# Use UHF response function for ROHF because in second order solver uhf
# response function is called to compute ROHF orbital hessian
rohf.ROHF.gen_response = _gen_uhf_response
dhf.DHF.gen_response = _gen_dhf_response



#!/usr/bin/env python
# Copyright 2014-2024 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

import numpy
from pyscf import lib

def _take_4d(a, indices):
    a_shape = a.shape
    ranges = []
    for i, s in enumerate(indices):
        if s is None:
            idx = numpy.arange(a_shape[i], dtype=numpy.int32)
        else:
            idx = numpy.asarray(s, dtype=numpy.int32)
            idx[idx < 0] += a_shape[i]
        ranges.append(idx)
    idx = ranges[0][:,None] * a_shape[1] + ranges[1]
    idy = ranges[2][:,None] * a_shape[3] + ranges[3]
    a = a.reshape(a_shape[0]*a_shape[1], a_shape[2]*a_shape[3])
    out = lib.take_2d(a, idx.ravel(), idy.ravel())
    return out.reshape([len(s) for s in ranges])

def _takebak_4d(out, a, indices):
    out_shape = out.shape
    a_shape = a.shape
    ranges = []
    for i, s in enumerate(indices):
        if s is None:
            idx = numpy.arange(a_shape[i], dtype=numpy.int32)
        else:
            idx = numpy.asarray(s, dtype=numpy.int32)
            idx[idx < 0] += out_shape[i]
        assert (len(idx) == a_shape[i])
        ranges.append(idx)
    idx = ranges[0][:,None] * out_shape[1] + ranges[1]
    idy = ranges[2][:,None] * out_shape[3] + ranges[3]
    nx = idx.size
    ny = idy.size
    out = out.reshape(out_shape[0]*out_shape[1], out_shape[2]*out_shape[3])
    lib.takebak_2d(out, a.reshape(nx,ny), idx.ravel(), idy.ravel())
    return out

def _take_5d(a, indices):
    a_shape = a.shape
    a = a.reshape((a_shape[0]*a_shape[1],) + a_shape[2:])
    indices = (None,) + indices[2:]
    return _take_4d(a, indices)

def _takebak_5d(out, a, indices):
    a_shape = a.shape
    out_shape = out.shape
    a = a.reshape((a_shape[0]*a_shape[1],) + a_shape[2:])
    out = out.reshape((out_shape[0]*out_shape[1],) + out_shape[2:])
    indices = (None,) + indices[2:]
    return _takebak_4d(out, a, indices)



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Zhi-Hao Cui <zhcui0408@gmail.com>
#

"""
Restricted DFT+U with kpoint sampling.
Based on KRHF routine.

Refs: PRB, 1998, 57, 1505.
"""

import itertools as it
import numpy as np
import scipy.linalg as la
from functools import reduce

from pyscf import lib
from pyscf.lib import logger
from pyscf import __config__
from pyscf.pbc.dft import krks
from pyscf.data.nist import HARTREE2EV
from pyscf import lo
from pyscf.lo import iao
from pyscf.pbc import gto as pgto

def get_veff(ks, cell=None, dm=None, dm_last=0, vhf_last=0, hermi=1,
             kpts=None, kpts_band=None):
    """
    Coulomb + XC functional + Hubbard U terms.

    .. note::
        This is a replica of pyscf.dft.rks.get_veff with kpts added.
        This function will change the ks object.

    Args:
        ks : an instance of :class:`RKS`
            XC functional are controlled by ks.xc attribute.  Attribute
            ks.grids might be initialized.
        dm : ndarray or list of ndarrays
            A density matrix or a list of density matrices

    Returns:
        Veff : ``(nkpts, nao, nao)`` or ``(*, nkpts, nao, nao)`` ndarray
        Veff = J + Vxc + V_U.
    """
    if cell is None: cell = ks.cell
    if dm is None: dm = ks.make_rdm1()
    if kpts is None: kpts = ks.kpts

    # J + V_xc
    vxc = krks.get_veff(ks, cell, dm, dm_last=dm_last, vhf_last=vhf_last,
                        hermi=hermi, kpts=kpts, kpts_band=kpts_band)
    vxc = _add_Vhubbard(vxc, ks, dm, kpts)
    return vxc

def _add_Vhubbard(vxc, ks, dm, kpts):
    '''Add Hubbard U to Vxc matrix inplace.
    '''
    C_ao_lo = ks.C_ao_lo
    ovlp = ks.get_ovlp()
    nkpts = len(kpts)
    nlo = C_ao_lo.shape[-1]

    rdm1_lo  = np.zeros((nkpts, nlo, nlo), dtype=np.complex128)
    for k in range(nkpts):
        C_inv = np.dot(C_ao_lo[k].conj().T, ovlp[k])
        rdm1_lo[k] = mdot(C_inv, dm[k], C_inv.conj().T)

    is_ibz = hasattr(kpts, "kpts_ibz")
    if is_ibz:
        rdm1_lo_0 = kpts.dm_at_ref_cell(rdm1_lo)

    alphas = ks.alpha
    if not hasattr(alphas, '__len__'): # not a list or tuple
        alphas = [alphas] * len(ks.U_idx)

    E_U = 0.0
    weight = getattr(kpts, "weights_ibz", np.repeat(1.0/nkpts, nkpts))
    logger.info(ks, "-" * 79)
    with np.printoptions(precision=5, suppress=True, linewidth=1000):
        for idx, val, lab, alpha in zip(ks.U_idx, ks.U_val, ks.U_lab, alphas):
            lab_string = " "
            for l in lab:
                lab_string += "%9s" %(l.split()[-1])
            lab_sp = lab[0].split()
            logger.info(ks, "local rdm1 of atom %s: ",
                        " ".join(lab_sp[:2]) + " " + lab_sp[2][:2])
            U_mesh = np.ix_(idx, idx)
            P_loc = 0.0
            for k in range(nkpts):
                S_k = ovlp[k]
                C_k = C_ao_lo[k][:, idx]
                P_k = rdm1_lo[k][U_mesh]
                E_U += weight[k] * (val * 0.5) * (P_k.trace() - np.dot(P_k, P_k).trace() * 0.5)
                vhub_loc = (np.eye(P_k.shape[-1]) - P_k) * (val * 0.5)
                if alpha is not None:
                    # The alpha perturbation is only applied to the linear term of
                    # the local density.
                    E_U += weight[k] * alpha * P_k.trace()
                    vhub_loc += np.eye(P_k.shape[-1]) * alpha
                SC = np.dot(S_k, C_k)
                vhub_loc = SC.dot(vhub_loc).dot(SC.conj().T)
                if vxc.dtype == np.float64:
                    vhub_loc = vhub_loc.real
                vxc[k] += vhub_loc
                if not is_ibz:
                    P_loc += P_k
            if is_ibz:
                P_loc = rdm1_lo_0[U_mesh].real
            else:
                P_loc = P_loc.real / nkpts
            logger.info(ks, "%s\n%s", lab_string, P_loc)
            logger.info(ks, "-" * 79)

    if E_U.real < 0.0 and all(np.asarray(ks.U_val) > 0):
        logger.warn(ks, "E_U (%g) is negative...", E_U.real)
    vxc = lib.tag_array(vxc, E_U=E_U)
    return vxc

def energy_elec(ks, dm_kpts=None, h1e_kpts=None, vhf=None):
    """
    Electronic energy for KRKSpU.
    """
    if h1e_kpts is None: h1e_kpts = ks.get_hcore(ks.cell, ks.kpts)
    if dm_kpts is None: dm_kpts = ks.make_rdm1()
    if vhf is None or getattr(vhf, 'ecoul', None) is None:
        vhf = ks.get_veff(ks.cell, dm_kpts)

    weight = getattr(ks.kpts, "weights_ibz",
                     np.array([1.0/len(h1e_kpts),]*len(h1e_kpts)))
    e1 = np.einsum('k,kij,kji', weight, h1e_kpts, dm_kpts)
    tot_e = e1 + vhf.ecoul + vhf.exc + vhf.E_U
    ks.scf_summary['e1'] = e1.real
    ks.scf_summary['coul'] = vhf.ecoul.real
    ks.scf_summary['exc'] = vhf.exc.real
    ks.scf_summary['E_U'] = vhf.E_U.real
    logger.debug(ks, 'E1 = %s  Ecoul = %s  Exc = %s  EU = %s', e1, vhf.ecoul,
                 vhf.exc, vhf.E_U)
    return tot_e.real, vhf.ecoul + vhf.exc + vhf.E_U

def set_U(ks, U_idx, U_val):
    """
    Regularize the U_idx and U_val to each atom,
    and set ks.U_idx, ks.U_val, ks.U_lab.
    """
    assert len(U_idx) == len(U_val)
    ks.U_val = []
    ks.U_idx = []
    ks.U_lab = []

    lo_labels = np.asarray(ks.cell.ao_labels())
    for i, idx in enumerate(U_idx):
        if isinstance(idx, str):
            lab_idx = ks.cell.search_ao_label(idx)
            labs = lo_labels[lab_idx]
            labs = zip(lab_idx, labs)
            for j, idxj in it.groupby(labs, key=lambda x: x[1].split()[0]):
                ks.U_idx.append(list(list(zip(*idxj))[0]))
                ks.U_val.append(U_val[i])
        else:
            ks.U_idx.append(idx)
            ks.U_val.append(U_val[i])

    ks.U_val = np.asarray(ks.U_val) / HARTREE2EV

    for idx, val in zip(ks.U_idx, ks.U_val):
        ks.U_lab.append(lo_labels[idx])

    if len(ks.U_idx) == 0:
        logger.warn(ks, "No sites specified for Hubbard U. "
                    "Please check if 'U_idx' is correctly specified")

def make_minao_lo(ks, minao_ref):
    """
    Construct minao local orbitals.
    """
    cell = ks.cell
    nao = cell.nao
    kpts = getattr(ks.kpts, "kpts_ibz", ks.kpts)
    nkpts = len(kpts)
    ovlp = ks.get_ovlp()
    C_ao_minao, labels = proj_ref_ao(cell, minao=minao_ref, kpts=kpts,
                                     return_labels=True)
    for k in range(nkpts):
        C_ao_minao[k] = lo.vec_lowdin(C_ao_minao[k], ovlp[k])

    C_ao_lo = np.zeros((nkpts, nao, nao), dtype=np.complex128)
    for idx, lab in zip(ks.U_idx, ks.U_lab):
        idx_minao = [i for i, l in enumerate(labels) if l in lab]
        assert len(idx_minao) == len(idx)
        C_ao_sub = C_ao_minao[:, :, idx_minao]
        C_ao_lo[:, :, idx] = C_ao_sub
    return C_ao_lo

def proj_ref_ao(mol, minao='minao', kpts=None, return_labels=False):
    """
    Get a set of reference AO spanned by the calculation basis.
    Not orthogonalized.

    Args:
        return_labels: if True, return the labels as well.
    """
    nkpts = len(kpts)
    pmol = iao.reference_mol(mol, minao)
    s1 = np.asarray(mol.pbc_intor('int1e_ovlp', hermi=1, kpts=kpts))
    s2 = np.asarray(pmol.pbc_intor('int1e_ovlp', hermi=1, kpts=kpts))
    s12 = np.asarray(pgto.cell.intor_cross('int1e_ovlp', mol, pmol, kpts=kpts))
    #s21 = np.swapaxes(s12, -1, -2).conj()
    C_ao_lo = np.zeros((nkpts, s1.shape[-1], s2.shape[-1]), dtype=np.complex128)
    for k in range(nkpts):
        s1cd_k = la.cho_factor(s1[k])
        #s2cd_k = la.cho_factor(s2[k])
        C_ao_lo[k] = la.cho_solve(s1cd_k, s12[k])

    if return_labels:
        labels = pmol.ao_labels()
        return C_ao_lo, labels
    else:
        return C_ao_lo

def mdot(*args):
    '''
    Compute the dot product of a list of arrays in a single function call.
    '''
    return reduce(np.dot, args)

def format_idx(idx_list):
    string = ''
    for k, g in it.groupby(enumerate(idx_list), lambda ix: ix[0] - ix[1]):
        g = list(g)
        if len(g) > 1:
            string += '%d-%d, '%(g[0][1], g[-1][1])
        else:
            string += '%d, '%(g[0][1])
    return string[:-2]

class KRKSpU(krks.KRKS):
    """
    RKSpU class adapted for PBCs with k-point sampling.
    """

    _keys = {"U_idx", "U_val", "C_ao_lo", "U_lab", 'alpha'}

    get_veff = get_veff
    energy_elec = energy_elec
    to_hf = lib.invalid_method('to_hf')

    def __init__(self, cell, kpts=np.zeros((1,3)), xc='LDA,VWN',
                 exxdiv=getattr(__config__, 'pbc_scf_SCF_exxdiv', 'ewald'),
                 U_idx=[], U_val=[], C_ao_lo='minao', minao_ref='MINAO', **kwargs):
        """
        DFT+U args:
            U_idx: can be
                   list of list: each sublist is a set of LO indices to add U.
                   list of string: each string is one kind of LO orbitals,
                                   e.g. ['Ni 3d', '1 O 2pz'], in this case,
                                   LO should be aranged as ao_labels order.
                   or a combination of these two.
            U_val: a list of effective U [in eV], i.e. U-J in Dudarev's DFT+U.
                   each U corresponds to one kind of LO orbitals, should have
                   the same length as U_idx.
            C_ao_lo: LO coefficients, can be
                     np.array, shape ((spin,), nkpts, nao, nlo),
                     string, in 'minao'.
            minao_ref: reference for minao orbitals, default is 'MINAO'.

        Attributes:
            U_idx: same as the input.
            U_val: effectiv U-J [in AU]
            C_ao_loc: np.array
            alpha: the perturbation [in AU] used to compute U in LR-cDFT.
                Refs: Cococcioni and de Gironcoli, PRB 71, 035105 (2005)
        """
        super(self.__class__, self).__init__(cell, kpts, xc=xc, exxdiv=exxdiv, **kwargs)

        set_U(self, U_idx, U_val)

        if isinstance(C_ao_lo, str):
            if C_ao_lo.upper() == 'MINAO':
                self.C_ao_lo = make_minao_lo(self, minao_ref)
            else:
                raise NotImplementedError
        else:
            self.C_ao_lo = np.asarray(C_ao_lo)
        if self.C_ao_lo.ndim == 4:
            self.C_ao_lo = self.C_ao_lo[0]

        # The perturbation (eV) used to compute U in LR-cDFT.
        self.alpha = None

    def dump_flags(self, verbose=None):
        super().dump_flags(verbose)
        log = logger.new_logger(self, verbose)
        if log.verbose >= logger.INFO:
            _print_U_info(self, log)
        return self

    def nuc_grad_method(self):
        raise NotImplementedError

def _print_U_info(mf, log):
    alphas = mf.alpha
    if not hasattr(alphas, '__len__'): # not a list or tuple
        alphas = [alphas] * len(mf.U_idx)
    log.info("-" * 79)
    log.info('U indices and values: ')
    for idx, val, lab, alpha in zip(mf.U_idx, mf.U_val, mf.U_lab, alphas):
        log.info('%6s [%.6g eV] ==> %-100s', format_idx(idx),
                    val * HARTREE2EV, "".join(lab))
        if alpha is not None:
            log.info('alpha for LR-cDFT %s (eV)', alpha*HARTREE2EV)
    log.info("-" * 79)

def linear_response_u(mf_plus_u, alphalist=(0.02, 0.05, 0.08)):
    '''
    Refs:
        [1] M. Cococcioni and S. de Gironcoli, Phys. Rev. B 71, 035105 (2005)
        [2] H. J. Kulik, M. Cococcioni, D. A. Scherlis, and N. Marzari, Phys. Rev. Lett. 97, 103001 (2006)
        [3] Heather J. Kulik, J. Chem. Phys. 142, 240901 (2015)
        [4] https://hjkgrp.mit.edu/tutorials/2011-05-31-calculating-hubbard-u/
        [5] https://hjkgrp.mit.edu/tutorials/2011-06-28-hubbard-u-multiple-sites/

    Args:
        alphalist :
            alpha parameters (in eV) are the displacements for the linear
            response calculations. For each alpha in this list, the DFT+U with
            U=u0+alpha, U=u0-alpha are evaluated. u0 is the U value from the
            reference mf_plus_u object, which will be treated as a standard DFT
            functional.
    '''
    is_ibz = hasattr(mf_plus_u.kpts, "kpts_ibz")
    if is_ibz:
        raise NotImplementedError

    assert isinstance(mf_plus_u, KRKSpU)
    assert len(mf_plus_u.U_idx) > 0
    if not mf_plus_u.converged:
        mf_plus_u.run()
    assert mf_plus_u.converged
    # The bare density matrix without adding U
    bare_dm = mf_plus_u.make_rdm1()

    mf = mf_plus_u.copy()
    log = logger.new_logger(mf)

    alphalist = np.asarray(alphalist)
    alphalist = np.append(-alphalist[::-1], alphalist)

    nkpts = len(mf.kpts)
    C_ao_lo = mf.C_ao_lo
    ovlp = mf.get_ovlp()
    C_inv = [[C_ao_lo[k][:,local_idx].conj().T.dot(ovlp[k]) for k in range(nkpts)]
             for local_idx in mf.U_idx]

    bare_occupancies = []
    final_occupancies = []
    for alpha in alphalist:
        # All in atomic unit
        mf.alpha = alpha / HARTREE2EV
        mf.kernel(dm0=bare_dm)
        local_occ = 0
        for c in C_inv:
            C_on_site = [c[k].dot(mf.mo_coeff[k]) for k in range(nkpts)]
            rdm1_lo = mf.make_rdm1(C_on_site, mf.mo_occ)
            local_occ += sum(x.trace().real for x in rdm1_lo)
        local_occ /= nkpts
        final_occupancies.append(local_occ)

        # The first iteration of SCF
        fock = mf.get_fock(dm=bare_dm)
        e, mo = mf.eig(fock, ovlp)
        local_occ = 0
        for c in C_inv:
            C_on_site = [c[k].dot(mo[k]) for k in range(nkpts)]
            rdm1_lo = mf.make_rdm1(C_on_site, mf.mo_occ)
            local_occ += sum(x.trace().real for x in rdm1_lo)
        local_occ /= nkpts
        bare_occupancies.append(local_occ)
        log.info('alpha=%f bare_occ=%g final_occ=%g',
                 alpha, bare_occupancies[-1], final_occupancies[-1])

    chi0, occ0 = np.polyfit(alphalist, bare_occupancies, deg=1)
    chif, occf = np.polyfit(alphalist, final_occupancies, deg=1)
    log.info('Line fitting chi0 = %f x + %f', chi0, occ0)
    log.info('Line fitting chif = %f x + %f', chif, occf)
    Uresp = 1./chi0 - 1./chif
    log.note('Uresp = %f, chi0 = %f, chif = %f', Uresp, chi0, chif)
    return Uresp



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Common density-matrix manipulations

import numpy as np
from pyscf import lib
from pyscf.mcscf.addons import StateAverageFCISolver
from pyscf.mcscf.addons import StateAverageMixFCISolver
from scipy import linalg

# DMRG solvers require special handling but dmrgscf is not always installed
try:
    from pyscf import dmrgscf
    DMRGCI = dmrgscf.DMRGCI
except ImportError:
    class DMRGCI :
        pass

def _get_fcisolver (mc, ci, state=0):
    '''Find the appropriate FCI solver, CI vector, and nelecas tuple to
    build single-state reduced density matrices. If state_average or
    state_average_mix is involved this takes a bit of work.

    The better solution, of course, is to edit StateAverage*FCI classes
    to have quick density-matrices-of-one-state API...
    '''
    nelecas = mc.nelecas
    nroots = getattr (mc.fcisolver, 'nroots', 1)
    fcisolver = mc.fcisolver
    solver_state_index = state
    if nroots>1: ci = ci[state]
    if isinstance (mc.fcisolver, StateAverageMixFCISolver):
        p0 = 0
        fcisolver = None
        for s in mc.fcisolver.fcisolvers:
            p1 = p0 + s.nroots
            if p0 <= state and state < p1:
                fcisolver = s
                nelecas = mc.fcisolver._get_nelec (s, nelecas)
                solver_state_index = state - p0
                break
            p0 = p1
        if fcisolver is None:
            raise RuntimeError ("Can't find FCI solver for state", state)
    elif isinstance (mc.fcisolver, StateAverageFCISolver):
        fcisolver = fcisolver.undo_state_average ()
    if isinstance (fcisolver, DMRGCI):
        ci = solver_state_index # DMRGCI takes state index in place of ci vector
    return fcisolver, ci, nelecas

def make_one_casdm1s (mc, ci, state=0):
    '''
    Construct the spin-separated active-space one-body reduced density
    matrix for a single state. This API is not consistently available
    in the StateAverageFCISolver functions without wasted effort (i.e.,
    without constructing the dms for all states and then discarding most
    of them
    '''
    ncas = mc.ncas
    fcisolver, ci, nelecas = _get_fcisolver (mc, ci, state=state)
    return fcisolver.make_rdm1s (ci, ncas, nelecas)

def make_one_casdm2 (mc, ci, state=0):
    '''
    Construct the spin-summed active-space two-body reduced density
    matrix for a single state. This API is not consistently available
    in the StateAverageFCISolver functions without wasted effort (i.e.,
    without constructing the dms for all states and then discarding most
    of them
    '''
    ncas = mc.ncas
    fcisolver, ci, nelecas = _get_fcisolver (mc, ci, state=state)
    try:
        casdm2 = fcisolver.make_rdm2 (ci, ncas, nelecas)
    except AttributeError:
        # Hail Mary: maybe the fcisolver class only has make_rdm12
        # but not make_rdm2 implemented?
        _, casdm2 = fcisolver.make_rdm12 (ci, ncas, nelecas)
    return casdm2


def dm2_cumulant (dm2, dm1s):
    '''
    Evaluate the spin-summed two-body cumulant reduced density
    matrix:

    cm2[p,q,r,s] = (dm2[p,q,r,s] - dm1[p,q]*dm1[r,s]
                       + dm1s[0][p,s]*dm1s[0][r,q]
                       + dm1s[1][p,s]*dm1s[1][r,q])

    Args:
        dm2 : ndarray of shape [norb,]*4
            Contains spin-summed 2-RDMs
        dm1s : ndarray (or compatible) of overall shape [2,norb,norb]
            Contains spin-separated 1-RDMs

    Returns:
        cm2 : ndarray of shape [norb,]*4
    '''

    dm1s = np.asarray (dm1s)
    if len (dm1s.shape) < 3:
        dm1 = dm1s.copy ()
        dm1s = dm1 / 2
        dm1s = np.stack ((dm1s, dm1s), axis=0)
    else:
        dm1 = dm1s[0] + dm1s[1]
    cm2  = dm2.copy ()
    cm2 -= np.multiply.outer (dm1, dm1)
    cm2 += np.multiply.outer (dm1s[0], dm1s[0]).transpose (0, 3, 2, 1)
    cm2 += np.multiply.outer (dm1s[1], dm1s[1]).transpose (0, 3, 2, 1)
    return cm2

def dm2s_cumulant (dm2s, dm1s):
    '''Evaluate the spin-summed two-body cumulant reduced density
    matrix:

    cm2s[0][p,q,r,s] = (dm2s[0][p,q,r,s] - dm1s[0][p,q]*dm1s[0][r,s]
                       + dm1s[0][p,s]*dm1s[0][r,q])
    cm2s[1][p,q,r,s] = (dm2s[1][p,q,r,s] - dm1s[0][p,q]*dm1s[1][r,s])
    cm2s[2][p,q,r,s] = (dm2s[2][p,q,r,s] - dm1s[1][p,q]*dm1s[1][r,s]
                       + dm1s[1][p,s]*dm1s[1][r,q])

    Args:
        dm2s : ndarray of shape [norb,]*4
            Contains spin-separated 2-RDMs
        dm1s : ndarray (or compatible) of overall shape [2,norb,norb]
            Contains spin-separated 1-RDMs

    Returns:
        cm2s : (cm2s[0], cms2[1], cm2s[2])
            ndarrays of shape [norb,]*4; contain spin components
            aa, ab, bb respectively
    '''
    dm1s = np.asarray (dm1s)
    if len (dm1s.shape) < 3:
        dm1 = dm1s.copy ()
        dm1s = dm1 / 2
        dm1s = np.stack ((dm1s, dm1s), axis=0)
    #cm2  = dm2 - np.einsum ('pq,rs->pqrs', dm1, dm1)
    #cm2 +=    0.5 * np.einsum ('ps,rq->pqrs', dm1, dm1)
    cm2s = [i.copy () for i in dm2s]
    cm2s[0] -= np.multiply.outer (dm1s[0], dm1s[0])
    cm2s[1] -= np.multiply.outer (dm1s[0], dm1s[1])
    cm2s[2] -= np.multiply.outer (dm1s[1], dm1s[1])
    cm2s[0] += np.multiply.outer (dm1s[0], dm1s[0]).transpose (0, 3, 2, 1)
    cm2s[2] += np.multiply.outer (dm1s[1], dm1s[1]).transpose (0, 3, 2, 1)
    return tuple (cm2s)

def casdm1s_to_dm1s (mc, casdm1s, mo_coeff=None, ncore=None, ncas=None):
    '''Generate AO-basis spin-separated 1-RDM from active space part.
    This is necessary because the StateAverageMCSCFSolver class doesn't
    have API for getting the AO-basis density matrix of a single state.

    Args:
        mc : object of CASCI or CASSCF class
        casdm1s : ndarray or compatible of shape (2,ncas,ncas)
            Active-space spin-separated 1-RDM

    Kwargs:
        ncore : integer
            Number of occupied inactive orbitals
        ncas : integer
            Number of active orbitals

    Returns:
        dm1s : ndarray of shape (2,nao,nao)
    '''
    if mo_coeff is None: mo_coeff=mc.mo_coeff
    if ncore is None: ncore=mc.ncore
    if ncas is None: ncas=mc.ncas
    mo_core = mo_coeff[:,:ncore]
    mo_cas = mo_coeff[:,ncore:][:,:ncas]
    moH_core = mo_core.conj ().T
    moH_cas = mo_cas.conj ().T

    casdm1s = np.asarray (casdm1s)
    dm1s_cas = np.dot (casdm1s, moH_cas)
    dm1s_cas = np.dot (mo_cas, dm1s_cas).transpose (1,0,2)
    dm1s_core = np.dot (mo_core, moH_core)
    dm1s = dm1s_cas + dm1s_core[None,:,:]

    # Tags for speeding up rho generators and DF fns
    no_coeff = mo_coeff[:,:ncore+ncas]
    no_coeff = np.stack ([no_coeff, no_coeff], axis=0)
    no_occ = np.zeros ((2,ncore+ncas), dtype=no_coeff.dtype)
    no_occ[:,:ncore] = 1.0
    no_cas = no_coeff[:,:,ncore:]
    for i in range (2):
        no_occ[i,ncore:], umat = linalg.eigh (-casdm1s[i])
        no_cas[i,:,:] = np.dot (no_cas[i,:,:], umat)
    no_occ[:,ncore:] *= -1
    dm1s = lib.tag_array (dm1s, mo_coeff=no_coeff, mo_occ=no_occ)

    return dm1s


def make_weighted_casdm1s(mc, ci=None, weights=None):
    '''Compute the weighted average 1-electron spin-separated CAS density.

    Args:
        mc : instance of class _PDFT

        ci : list of ndarrays of length nroots
            CI vectors should be from a converged CASSCF/CASCI calculation

        weights : ndarray of length nroots
            Weight for each state. If none, uses weights from SA-CASSCF
            calculation

    Returns:
        Weighted average of casdm1s
    '''
    if ci is None: ci = mc.ci
    if weights is None: weights = mc.weights

    # There might be a better way to construct all of them, but this should be
    # more cost-effective than what is currently in the _dms file.
    casdm1s_all = [make_one_casdm1s(mc, ci, state) for state in range(len(ci))]
    casdm1s_0 = np.tensordot(weights, casdm1s_all, axes=1)
    return tuple(casdm1s_0)

def make_weighted_casdm2(mc, ci=None, weights=None):
    '''Compute the weighted average 2-electron spin-summed CAS density.

    Args:
        mc : instance of class _PDFT

        ci : list of ndarrays of length nroots
            CI vectors should be from a converged CASSCF/CASCI calculation

        weights : ndarray of length nroots
            Weight for each state. If none, uses weights from SA-CASSCF
            calculation

    Returns:
        Weighted average of casdm2
    '''
    if ci is None: ci = mc.ci
    if weights is None: weights = mc.weights

    # There might be a better way to construct all of them, but this should be
    # more cost-effective than what is currently in the _dms file.
    casdm2_all = [make_one_casdm2(mc, ci, state) for state in range(len(ci))]
    return np.tensordot(weights, casdm2_all, axes=1)



# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Oliver J. Backhouse <olbackhouse@gmail.com>
#         Alejandro Santana-Bonilla <alejandro.santana_bonilla@kcl.ac.uk>
#         George H. Booth <george.booth@kcl.ac.uk>
#

import numpy as np
import ctypes
from pyscf import lib
from pyscf.agf2 import mpi_helper

libagf2 = lib.load_library('libagf2')


def cholesky_build(vv, vev, eps=1e-16):
    ''' Constructs the truncated auxiliaries from :attr:`vv` and :attr:`vev`.
        Performs a Cholesky decomposition via :func:`numpy.linalg.cholesky`,
        for a positive-definite or positive-semidefinite matrix. For the
        latter, the null space is removed.

        The :attr:`vv` matrix of :func:`build_se_part` can be positive-
        semidefinite when :attr:`gf_occ.naux` < :attr:`gf_occ.nphys` for
        the occupied self-energy, or :attr:`gf_vir.naux` < :attr:`gf_vir.nphys`
        for the virtual self-energy.
    '''

    nmo = vv.shape[0]

    try:
        b = np.linalg.cholesky(vv).T
    except np.linalg.LinAlgError:
        w, v = np.linalg.eigh(vv)
        w[w < eps] = eps
        vv_posdef = np.dot(np.dot(v, np.diag(w)), v.T.conj())
        b = np.linalg.cholesky(vv_posdef).T

    b_inv = np.linalg.inv(b)

    m = np.dot(np.dot(b_inv.T, vev), b_inv)
    e, c = np.linalg.eigh(m)
    c = np.dot(b.T, c[:nmo])

    return e, c


def build_mats_ragf2_incore(qeri, e_occ, e_vir, os_factor=1.0, ss_factor=1.0):
    ''' Wraps AGF2ee_vv_vev_islice
    '''

    fdrv = getattr(libagf2, 'AGF2ee_vv_vev_islice')

    assert qeri.ndim == 4
    nmo = qeri.shape[0]
    nocc = e_occ.size
    nvir = e_vir.size

    qeri = np.asarray(qeri, order='C')
    e_i = np.asarray(e_occ, order='C')
    e_a = np.asarray(e_vir, order='C')

    vv = np.zeros((nmo*nmo))
    vev = np.zeros((nmo*nmo))

    rank, size = mpi_helper.rank, mpi_helper.size
    istart = rank * nocc // size
    iend = nocc if rank == (size-1) else (rank+1) * nocc // size

    fdrv(qeri.ctypes.data_as(ctypes.c_void_p),
         e_i.ctypes.data_as(ctypes.c_void_p),
         e_a.ctypes.data_as(ctypes.c_void_p),
         ctypes.c_double(os_factor),
         ctypes.c_double(ss_factor),
         ctypes.c_int(nmo),
         ctypes.c_int(nocc),
         ctypes.c_int(nvir),
         ctypes.c_int(istart),
         ctypes.c_int(iend),
         vv.ctypes.data_as(ctypes.c_void_p),
         vev.ctypes.data_as(ctypes.c_void_p))

    vv = vv.reshape(nmo, nmo)
    vev = vev.reshape(nmo, nmo)

    mpi_helper.barrier()
    mpi_helper.allreduce_safe_inplace(vv)
    mpi_helper.allreduce_safe_inplace(vev)

    return vv, vev


def build_mats_ragf2_outcore(qeri, e_occ, e_vir, os_factor=1.0, ss_factor=1.0):
    ''' Python version of AGF2ee_vv_vev_islice to support outcore
    '''

    assert qeri.ndim == 4
    nmo = qeri.shape[0]
    nocc = e_occ.size

    vv = np.zeros((nmo, nmo))
    vev = np.zeros((nmo, nmo))

    fpos = os_factor + ss_factor
    fneg = -ss_factor

    eja = lib.direct_sum('j,a->ja', e_occ, -e_vir)
    eja = eja.ravel()

    for i in mpi_helper.nrange(nocc):
        xija = qeri[:,i].reshape(nmo, -1)
        xjia = qeri[:,:,i].reshape(nmo, -1)

        eija = eja + e_occ[i]

        vv = lib.dot(xija, xija.T, alpha=fpos, beta=1, c=vv)
        vv = lib.dot(xija, xjia.T, alpha=fneg, beta=1, c=vv)

        exija = xija * eija[None]

        vev = lib.dot(exija, xija.T, alpha=fpos, beta=1, c=vev)
        vev = lib.dot(exija, xjia.T, alpha=fneg, beta=1, c=vev)

    vv = vv.reshape(nmo, nmo)
    vev = vev.reshape(nmo, nmo)

    mpi_helper.barrier()
    mpi_helper.allreduce_safe_inplace(vv)
    mpi_helper.allreduce_safe_inplace(vev)

    return vv, vev


def build_mats_dfragf2_incore(qxi, qja, e_occ, e_vir, os_factor=1.0, ss_factor=1.0):
    ''' Wrapper for AGF2df_vv_vev_islice
    '''

    fdrv = getattr(libagf2, 'AGF2df_vv_vev_islice')

    naux = qxi.shape[0]
    nocc = e_occ.size
    nvir = e_vir.size
    nmo = qxi.size // (naux*nocc)
    assert qxi.size == (naux * nmo * nocc)
    assert qja.size == (naux * nocc * nvir)

    qxi = np.asarray(qxi, order='C')
    qja = np.asarray(qja, order='C')
    e_i = np.asarray(e_occ, order='C')
    e_a = np.asarray(e_vir, order='C')

    rank, size = mpi_helper.rank, mpi_helper.size

    vv = np.zeros((nmo*nmo))
    vev = np.zeros((nmo*nmo))

    start = rank * nocc // size
    end = nocc if rank == (size-1) else (rank+1) * nocc // size

    fdrv(qxi.ctypes.data_as(ctypes.c_void_p),
         qja.ctypes.data_as(ctypes.c_void_p),
         e_i.ctypes.data_as(ctypes.c_void_p),
         e_a.ctypes.data_as(ctypes.c_void_p),
         ctypes.c_double(os_factor),
         ctypes.c_double(ss_factor),
         ctypes.c_int(nmo),
         ctypes.c_int(nocc),
         ctypes.c_int(nvir),
         ctypes.c_int(naux),
         ctypes.c_int(start),
         ctypes.c_int(end),
         vv.ctypes.data_as(ctypes.c_void_p),
         vev.ctypes.data_as(ctypes.c_void_p))

    vv = vv.reshape(nmo, nmo)
    vev = vev.reshape(nmo, nmo)

    mpi_helper.barrier()
    mpi_helper.allreduce_safe_inplace(vv)
    mpi_helper.allreduce_safe_inplace(vev)

    return vv, vev


def build_mats_dfragf2_lowmem(qxi, qja, e_occ, e_vir, os_factor=1.0, ss_factor=1.0):
    ''' Wrapper for AGF2df_vv_vev_islice_lowmem
    '''

    fdrv = getattr(libagf2, 'AGF2df_vv_vev_islice_lowmem')

    naux = qxi.shape[0]
    nocc = e_occ.size
    nvir = e_vir.size
    nmo = qxi.size // (naux*nocc)
    assert qxi.size == (naux * nmo * nocc)
    assert qja.size == (naux * nocc * nvir)

    qxi = np.asarray(qxi, order='C')
    qja = np.asarray(qja, order='C')
    e_i = np.asarray(e_occ, order='C')
    e_a = np.asarray(e_vir, order='C')

    rank, size = mpi_helper.rank, mpi_helper.size

    vv = np.zeros((nmo*nmo))
    vev = np.zeros((nmo*nmo))

    start = rank * (nocc * nocc) // size
    end = nocc*nocc if rank == (size-1) else (rank+1) * (nocc*nocc) // size

    fdrv(qxi.ctypes.data_as(ctypes.c_void_p),
         qja.ctypes.data_as(ctypes.c_void_p),
         e_i.ctypes.data_as(ctypes.c_void_p),
         e_a.ctypes.data_as(ctypes.c_void_p),
         ctypes.c_double(os_factor),
         ctypes.c_double(ss_factor),
         ctypes.c_int(nmo),
         ctypes.c_int(nocc),
         ctypes.c_int(nvir),
         ctypes.c_int(naux),
         ctypes.c_int(start),
         ctypes.c_int(end),
         vv.ctypes.data_as(ctypes.c_void_p),
         vev.ctypes.data_as(ctypes.c_void_p))

    vv = vv.reshape(nmo, nmo)
    vev = vev.reshape(nmo, nmo)

    mpi_helper.barrier()
    mpi_helper.allreduce_safe_inplace(vv)
    mpi_helper.allreduce_safe_inplace(vev)

    return vv, vev


def build_mats_dfragf2_outcore(qxi, qja, e_occ, e_vir, os_factor=1.0, ss_factor=1.0):
    ''' Python version of AGF2df_vv_vev_islice to support outcore
    '''

    naux = qxi.shape[0]
    nocc = e_occ.size
    nvir = e_vir.size
    nmo = qxi.size // (naux*nocc)
    assert qxi.size == (naux * nmo * nocc)
    assert qja.size == (naux * nocc * nvir)

    vv = np.zeros((nmo, nmo))
    vev = np.zeros((nmo, nmo))

    fpos = os_factor + ss_factor
    fneg = -ss_factor

    eja = lib.direct_sum('j,a->ja', e_occ, -e_vir)
    eja = eja.ravel()

    buf = (np.zeros((nmo, nocc*nvir)), np.zeros((nmo*nocc, nvir)))

    for i in mpi_helper.nrange(nocc):
        qx = qxi.reshape(naux, nmo, nocc)[:,:,i]
        xija = lib.dot(qx.T, qja, c=buf[0])
        xjia = lib.dot(qxi.T, qja[:,i*nvir:(i+1)*nvir], c=buf[1])
        xjia = xjia.reshape(nmo, nocc*nvir)

        eija = eja + e_occ[i]

        vv = lib.dot(xija, xija.T, alpha=fpos, beta=1, c=vv)
        vv = lib.dot(xija, xjia.T, alpha=fneg, beta=1, c=vv)

        exija = xija * eija[None]

        vev = lib.dot(exija, xija.T, alpha=fpos, beta=1, c=vev)
        vev = lib.dot(exija, xjia.T, alpha=fneg, beta=1, c=vev)

    vv = vv.reshape(nmo, nmo)
    vev = vev.reshape(nmo, nmo)

    mpi_helper.barrier()
    mpi_helper.allreduce_safe_inplace(vv)
    mpi_helper.allreduce_safe_inplace(vev)

    return vv, vev


def build_mats_uagf2_incore(qeri, e_occ, e_vir, os_factor=1.0, ss_factor=1.0):
    ''' Wraps AGF2uee_vv_vev_islice
    '''

    fdrv = getattr(libagf2, 'AGF2uee_vv_vev_islice')

    assert qeri[0].ndim == qeri[1].ndim == 4
    nmo = qeri[0].shape[0]
    noa, nob = e_occ[0].size, e_occ[1].size
    nva, nvb = e_vir[0].size, e_vir[1].size

    qeri_a = np.asarray(qeri[0], order='C')
    qeri_b = np.asarray(qeri[1], order='C')
    e_i = np.asarray(e_occ[0], order='C')
    e_I = np.asarray(e_occ[1], order='C')
    e_a = np.asarray(e_vir[0], order='C')
    e_A = np.asarray(e_vir[1], order='C')

    vv = np.zeros((nmo*nmo))
    vev = np.zeros((nmo*nmo))

    rank, size = mpi_helper.rank, mpi_helper.size
    istart = rank * noa // size
    iend = noa if rank == (size-1) else (rank+1) * noa // size

    fdrv(qeri_a.ctypes.data_as(ctypes.c_void_p),
         qeri_b.ctypes.data_as(ctypes.c_void_p),
         e_i.ctypes.data_as(ctypes.c_void_p),
         e_I.ctypes.data_as(ctypes.c_void_p),
         e_a.ctypes.data_as(ctypes.c_void_p),
         e_A.ctypes.data_as(ctypes.c_void_p),
         ctypes.c_double(os_factor),
         ctypes.c_double(ss_factor),
         ctypes.c_int(nmo),
         ctypes.c_int(noa),
         ctypes.c_int(nob),
         ctypes.c_int(nva),
         ctypes.c_int(nvb),
         ctypes.c_int(istart),
         ctypes.c_int(iend),
         vv.ctypes.data_as(ctypes.c_void_p),
         vev.ctypes.data_as(ctypes.c_void_p))

    vv = vv.reshape(nmo, nmo)
    vev = vev.reshape(nmo, nmo)

    mpi_helper.barrier()
    mpi_helper.allreduce_safe_inplace(vv)
    mpi_helper.allreduce_safe_inplace(vev)

    return vv, vev


def build_mats_uagf2_outcore(qeri, e_occ, e_vir, os_factor=1.0, ss_factor=1.0):
    ''' Python version of AGF2uee_vv_vev_islice to support outcore
    '''

    assert qeri[0].ndim == qeri[1].ndim == 4
    nmo = qeri[0].shape[0]
    # noa, nob = e_occ[0].size, e_occ[1].size
    # nva, nvb = e_vir[0].size, e_vir[1].size
    noa = e_occ[0].size

    vv = np.zeros((nmo, nmo))
    vev = np.zeros((nmo, nmo))

    fposa = ss_factor
    fnega = -ss_factor
    fposb = os_factor

    eja_a = lib.direct_sum('j,a->ja', e_occ[0], -e_vir[0]).ravel()
    eja_b = lib.direct_sum('j,a->ja', e_occ[1], -e_vir[1]).ravel()

    for i in mpi_helper.nrange(noa):
        xija_aa = qeri[0][:,i].reshape(nmo, -1)
        xija_ab = qeri[1][:,i].reshape(nmo, -1)
        xjia_aa = qeri[0][:,:,i].reshape(nmo, -1)

        eija_aa = eja_a + e_occ[0][i]
        eija_ab = eja_b + e_occ[0][i]

        vv = lib.dot(xija_aa, xija_aa.T, alpha=fposa, beta=1, c=vv)
        vv = lib.dot(xija_aa, xjia_aa.T, alpha=fnega, beta=1, c=vv)
        vv = lib.dot(xija_ab, xija_ab.T, alpha=fposb, beta=1, c=vv)

        exija_aa = xija_aa * eija_aa[None]
        exija_ab = xija_ab * eija_ab[None]

        vev = lib.dot(exija_aa, xija_aa.T, alpha=fposa, beta=1, c=vev)
        vev = lib.dot(exija_aa, xjia_aa.T, alpha=fnega, beta=1, c=vev)
        vev = lib.dot(exija_ab, xija_ab.T, alpha=fposb, beta=1, c=vev)

    vv = vv.reshape(nmo, nmo)
    vev = vev.reshape(nmo, nmo)

    mpi_helper.barrier()
    mpi_helper.allreduce_safe_inplace(vv)
    mpi_helper.allreduce_safe_inplace(vev)

    return vv, vev


def build_mats_dfuagf2_incore(qxi, qja, e_occ, e_vir, os_factor=1.0, ss_factor=1.0):
    ''' Wrapper for AGF2udf_vv_vev_islice
    '''

    fdrv = getattr(libagf2, 'AGF2udf_vv_vev_islice')

    naux = qxi[0].shape[0]
    noa, nob = e_occ[0].size, e_occ[1].size
    nva, nvb = e_vir[0].size, e_vir[1].size
    nmo = qxi[0].size // (naux*noa)
    assert qxi[0].size == (naux * nmo * noa)
    assert qja[0].size == (naux * noa * nva)
    assert qja[1].size == (naux * nob * nvb)

    qxi_a, qxi_b = qxi
    qja_a, qja_b = qja

    qxi = np.asarray(qxi_a, order='C')
    qja = np.asarray(qja_a, order='C')
    qJA = np.asarray(qja_b, order='C')
    e_i = np.asarray(e_occ[0], order='C')
    e_I = np.asarray(e_occ[1], order='C')
    e_a = np.asarray(e_vir[0], order='C')
    e_A = np.asarray(e_vir[1], order='C')

    vv = np.zeros((nmo*nmo))
    vev = np.zeros((nmo*nmo))

    rank, size = mpi_helper.rank, mpi_helper.size
    istart = rank * noa // size
    iend = noa if rank == (size-1) else (rank+1) * noa // size

    fdrv(qxi.ctypes.data_as(ctypes.c_void_p),
         qja.ctypes.data_as(ctypes.c_void_p),
         qJA.ctypes.data_as(ctypes.c_void_p),
         e_i.ctypes.data_as(ctypes.c_void_p),
         e_I.ctypes.data_as(ctypes.c_void_p),
         e_a.ctypes.data_as(ctypes.c_void_p),
         e_A.ctypes.data_as(ctypes.c_void_p),
         ctypes.c_double(os_factor),
         ctypes.c_double(ss_factor),
         ctypes.c_int(nmo),
         ctypes.c_int(noa),
         ctypes.c_int(nob),
         ctypes.c_int(nva),
         ctypes.c_int(nvb),
         ctypes.c_int(naux),
         ctypes.c_int(istart),
         ctypes.c_int(iend),
         vv.ctypes.data_as(ctypes.c_void_p),
         vev.ctypes.data_as(ctypes.c_void_p))

    vv = vv.reshape(nmo, nmo)
    vev = vev.reshape(nmo, nmo)

    mpi_helper.barrier()
    mpi_helper.allreduce_safe_inplace(vv)
    mpi_helper.allreduce_safe_inplace(vev)

    return vv, vev


def build_mats_dfuagf2_lowmem(qxi, qja, e_occ, e_vir, os_factor=1.0, ss_factor=1.0):
    ''' Wrapper for AGF2udf_vv_vev_islice_lowmem
    '''

    fdrv = getattr(libagf2, 'AGF2udf_vv_vev_islice_lowmem')

    naux = qxi[0].shape[0]
    noa, nob = e_occ[0].size, e_occ[1].size
    nva, nvb = e_vir[0].size, e_vir[1].size
    nmo = qxi[0].size // (naux*noa)
    assert qxi[0].size == (naux * nmo * noa)
    assert qja[0].size == (naux * noa * nva)
    assert qja[1].size == (naux * nob * nvb)

    qxi_a, qxi_b = qxi
    qja_a, qja_b = qja

    qxi = np.asarray(qxi_a, order='C')
    qja = np.asarray(qja_a, order='C')
    qJA = np.asarray(qja_b, order='C')
    e_i = np.asarray(e_occ[0], order='C')
    e_I = np.asarray(e_occ[1], order='C')
    e_a = np.asarray(e_vir[0], order='C')
    e_A = np.asarray(e_vir[1], order='C')

    vv = np.zeros((nmo*nmo))
    vev = np.zeros((nmo*nmo))

    rank, size = mpi_helper.rank, mpi_helper.size
    nomax = max(noa, nob)
    start = rank * (noa*nomax) // size
    end = (noa*nomax) if rank == (size-1) else (rank+1) * (noa*nomax) // size

    fdrv(qxi.ctypes.data_as(ctypes.c_void_p),
         qja.ctypes.data_as(ctypes.c_void_p),
         qJA.ctypes.data_as(ctypes.c_void_p),
         e_i.ctypes.data_as(ctypes.c_void_p),
         e_I.ctypes.data_as(ctypes.c_void_p),
         e_a.ctypes.data_as(ctypes.c_void_p),
         e_A.ctypes.data_as(ctypes.c_void_p),
         ctypes.c_double(os_factor),
         ctypes.c_double(ss_factor),
         ctypes.c_int(nmo),
         ctypes.c_int(noa),
         ctypes.c_int(nob),
         ctypes.c_int(nva),
         ctypes.c_int(nvb),
         ctypes.c_int(naux),
         ctypes.c_int(start),
         ctypes.c_int(end),
         vv.ctypes.data_as(ctypes.c_void_p),
         vev.ctypes.data_as(ctypes.c_void_p))

    vv = vv.reshape(nmo, nmo)
    vev = vev.reshape(nmo, nmo)

    mpi_helper.barrier()
    mpi_helper.allreduce_safe_inplace(vv)
    mpi_helper.allreduce_safe_inplace(vev)

    return vv, vev


def build_mats_dfuagf2_outcore(qxi, qja, e_occ, e_vir, os_factor=1.0, ss_factor=1.0):
    ''' Python version of AGF2udf_vv_vev_islice to support outcore
    '''

    naux = qxi[0].shape[0]
    noa, nob = e_occ[0].size, e_occ[1].size
    nva, nvb = e_vir[0].size, e_vir[1].size
    nmo = qxi[0].size // (naux*noa)
    assert qxi[0].size == (naux * nmo * noa)
    assert qja[0].size == (naux * noa * nva)
    assert qja[1].size == (naux * nob * nvb)

    qxi_a, qxi_b = qxi
    qja_a, qja_b = qja

    vv = np.zeros((nmo, nmo))
    vev = np.zeros((nmo, nmo))

    fposa = ss_factor
    fnega = -ss_factor
    fposb = os_factor

    eja_a = lib.direct_sum('j,a->ja', e_occ[0], -e_vir[0]).ravel()
    eja_b = lib.direct_sum('j,a->ja', e_occ[1], -e_vir[1]).ravel()

    buf = (np.zeros((nmo, noa*nva)),
           np.zeros((nmo, nob*nvb)),
           np.zeros((nmo*noa, nva)))

    for i in mpi_helper.nrange(noa):
        qx_a = qxi_a.reshape(naux, nmo, noa)[:,:,i]
        xija_aa = lib.dot(qx_a.T, qja_a, c=buf[0])
        xija_ab = lib.dot(qx_a.T, qja_b, c=buf[1])
        xjia_aa = lib.dot(qxi_a.T, qja_a[:,i*nva:(i+1)*nva], c=buf[2])
        xjia_aa = xjia_aa.reshape(nmo, -1)

        eija_aa = eja_a + e_occ[0][i]
        eija_ab = eja_b + e_occ[0][i]

        vv = lib.dot(xija_aa, xija_aa.T, alpha=fposa, beta=1, c=vv)
        vv = lib.dot(xija_aa, xjia_aa.T, alpha=fnega, beta=1, c=vv)
        vv = lib.dot(xija_ab, xija_ab.T, alpha=fposb, beta=1, c=vv)

        exija_aa = xija_aa * eija_aa[None]
        exija_ab = xija_ab * eija_ab[None]

        vev = lib.dot(exija_aa, xija_aa.T, alpha=fposa, beta=1, c=vev)
        vev = lib.dot(exija_aa, xjia_aa.T, alpha=fnega, beta=1, c=vev)
        vev = lib.dot(exija_ab, xija_ab.T, alpha=fposb, beta=1, c=vev)

    vv = vv.reshape(nmo, nmo)
    vev = vev.reshape(nmo, nmo)

    mpi_helper.barrier()
    mpi_helper.allreduce_safe_inplace(vv)
    mpi_helper.allreduce_safe_inplace(vev)

    return vv, vev


def get_blksize(max_memory_total, *sizes):
    ''' Gets a block size such that the sum of the product of
        :attr:`sizes` with :attr:`blksize` is less than avail
        memory.

        If multiple tuples are provided, the maximum is used.
    '''

    if isinstance(sizes[0], tuple):
        sum_of_sizes = max([sum(x) for x in sizes])
    else:
        sum_of_sizes = sum(sizes)

    mem_avail = max_memory_total - lib.current_memory()[0]
    mem_avail *= 8e6 # MB -> bits
    sum_of_sizes *= 64 # 64 bits -> bits

    return int(mem_avail / sum_of_sizes)



#!/usr/bin/env python
# Copyright 2014-2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Non-relativistic Unrestricted Kohn-Sham
'''

from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import uhf_symm
from pyscf.dft import uks
from pyscf.dft import rks


class SymAdaptedUKS(rks.KohnShamDFT, uhf_symm.UHF):
    ''' Restricted Kohn-Sham '''
    def __init__(self, mol, xc='LDA,VWN'):
        uhf_symm.UHF.__init__(self, mol)
        rks.KohnShamDFT.__init__(self, xc)

    def dump_flags(self, verbose=None):
        uhf_symm.UHF.dump_flags(self, verbose)
        rks.KohnShamDFT.dump_flags(self, verbose)
        return self

    get_veff = uks.get_veff
    get_vsap = uks.get_vsap
    energy_elec = uks.energy_elec

    init_guess_by_vsap = rks.init_guess_by_vsap

    def nuc_grad_method(self):
        from pyscf.grad import uks
        return uks.Gradients(self)

    to_gpu = lib.to_gpu

UKS = SymAdaptedUKS


if __name__ == '__main__':
    from pyscf import gto
    mol = gto.Mole()
    mol.verbose = 7
    mol.output = '/dev/null'

    mol.atom.extend([['He', (0.,0.,0.)], ])
    mol.basis = { 'He': 'cc-pvdz'}
    mol.symmetry = 1
    mol.build()

    m = UKS(mol)
    m.xc = 'b3lyp'
    print(m.scf())  # -2.89992555753



#!/usr/bin/env python
# Copyright 2022-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Xing Zhang <zhangxing.nju@gmail.com>
#

from functools import reduce
import numpy as np

from pyscf import lib
from pyscf.lib import logger
from pyscf.pbc.lib import ktensor
from pyscf.pbc.lib.kpts import MORotationMatrix, KQuartets

from pyscf.pbc.df import GDF, RSGDF
from pyscf.pbc.mp.kmp2 import (
    get_nocc,
    padded_mo_coeff,
    padding_k_idx,
)
from pyscf.pbc.cc import kintermediates_rhf_ksymm as imdk
from pyscf.pbc.cc.kccsd_rhf import (
    RCCSD,
    _get_epq,
    _init_df_eris,
)

einsum = lib.einsum

def update_amps(cc, t1, t2, eris):
    kpts = cc.kpts
    kqrts = cc.kqrts
    rmat = cc.rmat
    kconserv = cc.khelper.kconserv

    nkpts, nocc, nvir = t1.shape
    fock = eris.fock
    mo_e_o = [e[:nocc] for e in eris.mo_energy]
    mo_e_v = [e[nocc:] for e in eris.mo_energy]

    # Get location of padded elements in occupied and virtual space
    nonzero_opadding, nonzero_vpadding = padding_k_idx(cc, kind="split")

    ki_ibz_bz = kpts.ibz2bz[np.arange(kpts.nkpts_ibz)]
    fov = fock[:, :nocc, nocc:]
    kconserv = cc.khelper.kconserv

    Foo = imdk.cc_Foo(kpts, kqrts, t1, t2, eris, rmat)
    Fvv = imdk.cc_Fvv(kpts, kqrts, t1, t2, eris, rmat)
    Fov = imdk.cc_Fov(kpts, kqrts, t1, t2, eris, rmat)
    Loo = imdk.Loo(kpts, kqrts, t1, t2, eris, rmat)
    Lvv = imdk.Lvv(kpts, kqrts, t1, t2, eris, rmat)

    Fov = Fov.todense()

    # Move energy terms to the other side
    for ki_ibz in range(kpts.nkpts_ibz):
        ki = kpts.ibz2bz[ki_ibz]
        Foo[ki][np.diag_indices(nocc)] -= mo_e_o[ki]
        Fvv[ki][np.diag_indices(nvir)] -= mo_e_v[ki]
        Loo[ki][np.diag_indices(nocc)] -= mo_e_o[ki]
        Lvv[ki][np.diag_indices(nvir)] -= mo_e_v[ki]

    t1new = ktensor.empty_like(t1)
    t1 = t1.todense()
    t2new = ktensor.empty_like(t2)
    t2 = t2.todense()

    # T1 equation
    for ka_ibz in range(kpts.nkpts_ibz):
        ki = ka = kpts.ibz2bz[ka_ibz]
        t1new[ka]  = fov[ka].conj()
        t1new[ka] += -2. * einsum('kc,ka,ic->ia', fov[ki], t1[ka], t1[ki])
        t1new[ka] += einsum('ac,ic->ia', Fvv[ka], t1[ki])
        t1new[ka] += -einsum('ki,ka->ia', Foo[ki], t1[ka])

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ki, kk, kc, kd = kq
        ka = ki

        Svovv = 2 * eris.vovv[ka, kk, kc] - eris.vovv[ka, kk, kd].transpose(0, 1, 3, 2)
        tau_term_1 = t2[ki, kk, kc].copy()
        if ki == kc and kk == kd:
            tau_term_1 += einsum('ic,kd->ikcd', t1[ki], t1[kk])
        fock = einsum('akcd,ikcd->ia', Svovv, tau_term_1)
        for _, iop in kqrts.loop_stabilizer(i):
            rmat_oo = rmat.oo[ka][iop]
            rmat_vv = rmat.vv[ka][iop]
            t1new[ka] += einsum('ia,im,ae->me', fock, rmat_oo, rmat_vv.conj())

    for i, kq in enumerate(kqrts.kqrts_ibz):
        kk, kl, ki, kd = kq
        ka = ki
        Sooov = 2 * eris.ooov[kk, kl, ki] - eris.ooov[kl, kk, ki].transpose(1, 0, 2, 3)
        tau_term_1 = t2[kk, kl, ka].copy()
        if kk == ka and kl == kc:
            tau_term_1 += einsum('ka,lc->klac', t1[ka], t1[kc])
        fock = -einsum('klic,klac->ia', Sooov, tau_term_1)

        op_group = kqrts.stars_ops[i]
        ka_prim = kpts.k2opk[ka, op_group]
        mask = np.isin(ka_prim, ki_ibz_bz)
        for iop, ka_p in zip(op_group[mask], ka_prim[mask]):
            rmat_oo = rmat.oo[ka][iop]
            rmat_vv = rmat.vv[ka][iop]
            t1new[ka_p] += einsum('ia,im,ae->me', fock, rmat_oo, rmat_vv.conj())

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ki, kk, ka, kc = kq
        if ka == ki and kk == kc:
            tau_term = 2 * t2[kk, ki, kk] - t2[ki, kk, kk].transpose(1, 0, 2, 3)
            if ki == kk:
                tau_term += einsum('ic,ka->kica', t1[ki], t1[ka])

            fock = einsum('kc,kica->ia', Fov[kc], tau_term)
            fock += einsum('akic,kc->ia', 2 * eris.voov[ka, kk, ki], t1[kc])
            fock += einsum('kaic,kc->ia', -eris.ovov[kk, ka, ki], t1[kc])

            for _, iop in kqrts.loop_stabilizer(i):
                rmat_oo = rmat.oo[ka][iop]
                rmat_vv = rmat.vv[ka][iop]
                t1new[ka] += einsum('ia,im,ae->me', fock, rmat_oo, rmat_vv.conj())

    for ki_ibz in range(kpts.nkpts_ibz):
        ka = ki = kpts.ibz2bz[ki_ibz]
        # Remove zero/padded elements from denominator
        eia = _get_epq([0,nocc,ki,mo_e_o,nonzero_opadding],
                       [0,nvir,ka,mo_e_v,nonzero_vpadding],
                       fac=[1.0,-1.0])
        t1new[ki] /= eia


    # T2 equation
    Loo = Loo.todense()
    Lvv = Lvv.todense()

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ki, kj, ka, kb = kq
        t2new[ki, kj, ka] = eris.oovv[ki, kj, ka].conj()

    mem_now = lib.current_memory()[0]
    if (cc.incore_complete or
        _memory_4d(cc, [nocc,]*4) + mem_now < cc.max_memory * .9):
        Woooo = imdk.cc_Woooo(kpts, kqrts, t1, t2, eris, rmat)
    else:
        metadata = {'kpts': kpts, 'kqrts': kqrts, 'rmat': rmat,
                    'label': 'oooo', 'trans': 'ccnn',
                    'incore': False}
        Woooo = ktensor.empty([nocc,]*4, dtype=t1.dtype, metadata=metadata)
        Woooo = imdk.cc_Woooo(kpts, kqrts, t1, t2, eris, rmat, Woooo)

    mem_now = lib.current_memory()[0]
    if (not cc.ktensor_direct and
        _memory_4d(cc, [nocc,]*4, False) + mem_now < cc.max_memory * .9):
        Woooo = Woooo.todense()

    def _t2_oooo(ki,kj,ka,kb):
        t2new_tmp = 0
        for kl in range(nkpts):
            kk = kconserv[kj, kl, ki]
            tau_term = t2[kk, kl, ka].copy()
            if kl == kb and kk == ka:
                tau_term += einsum('ic,jd->ijcd', t1[ka], t1[kb])
            t2new_tmp += 0.5 * einsum('klij,klab->ijab', Woooo[kk, kl, ki], tau_term)
        return t2new_tmp

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ki, kj, ka, kb = kq
        t2new_tmp = _t2_oooo(ki,kj,ka,kb)
        t2new[ki, kj, ka] += t2new_tmp
        if lib.isin_1d((kj,ki,kb,ka), kqrts.kqrts_ibz):
            t2new[kj, ki, kb] += t2new_tmp.transpose(1, 0, 3, 2)
        else:
            t2new_tmp = _t2_oooo(kj,ki,kb,ka)
            t2new[ki, kj, ka] += t2new_tmp.transpose(1, 0, 3, 2)

    Woooo = None

    add_vvvv_(cc, t2new, t1, t2, eris)

    def _t2_voov1(ki,kj,ka,kb):
        t2new_tmp = einsum('ac,ijcb->ijab', Lvv[ka], t2[ki, kj, ka])
        t2new_tmp += einsum('ki,kjab->ijab', -Loo[ki], t2[ki, kj, ka])

        kc = kj
        tmp2 = np.asarray(eris.vovv[kc, ki, kb]).transpose(3, 2, 1, 0).conj() \
               - einsum('kbic,ka->abic', eris.ovov[ka, kb, ki], t1[ka])
        t2new_tmp += einsum('abic,jc->ijab', tmp2, t1[kj])

        kk = kb
        tmp2 = np.asarray(eris.ooov[kj, ki, kk]).transpose(3, 2, 1, 0).conj() \
               + einsum('akic,jc->akij', eris.voov[ka, kk, ki], t1[kj])
        t2new_tmp -= einsum('akij,kb->ijab', tmp2, t1[kb])
        return t2new_tmp

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ki, kj, ka, kb = kq
        t2new_tmp = _t2_voov1(ki,kj,ka,kb)
        t2new[ki, kj, ka] += t2new_tmp
        if lib.isin_1d((kj,ki,kb,ka), kqrts.kqrts_ibz):
            t2new[kj, ki, kb] += t2new_tmp.transpose(1, 0, 3, 2)
        else:
            t2new_tmp = _t2_voov1(kj,ki,kb,ka)
            t2new[ki, kj, ka] += t2new_tmp.transpose(1, 0, 3, 2)

    mem_now = lib.current_memory()[0]
    if (cc.incore_complete or
        _memory_4d(cc, [nocc,nocc,nvir,nvir])*2 + mem_now < cc.max_memory*.9):
        Wvoov = imdk.cc_Wvoov(kpts, kqrts, t1, t2, eris, rmat)
        Wvovo = imdk.cc_Wvovo(kpts, kqrts, t1, t2, eris, rmat)
    else:
        metadata = {'kpts': kpts, 'kqrts': kqrts, 'rmat': rmat,
                    'trans': 'ccnn', 'incore': False}
        Wvoov = ktensor.empty([nvir,nocc,nocc,nvir], dtype=t1.dtype,
                              metadata={**metadata, 'label':'voov'})
        Wvovo = ktensor.empty([nvir,nocc,nvir,nocc], dtype=t1.dtype,
                              metadata={**metadata, 'label':'vovo'})
        Wvoov = imdk.cc_Wvoov(kpts, kqrts, t1, t2, eris, rmat, Wvoov)
        Wvovo = imdk.cc_Wvovo(kpts, kqrts, t1, t2, eris, rmat, Wvovo)

    mem_now = lib.current_memory()[0]
    if (not cc.ktensor_direct and
        _memory_4d(cc, [nocc,nocc,nvir,nvir], False)*2 + mem_now < cc.max_memory*.9):
        Wvoov = Wvoov.todense()
        Wvovo = Wvovo.todense()

    def _t2_voov2(ki,kj,ka,kb):
        t2new_tmp = 0
        for kk in range(nkpts):
            kc = kconserv[ka, ki, kk]
            tmp_voov = 2. * Wvoov[ka, kk, ki] - Wvovo[ka, kk, kc].transpose(0, 1, 3, 2)
            t2new_tmp += einsum('akic,kjcb->ijab', tmp_voov, t2[kk, kj, kc])

            kc = kconserv[ka, ki, kk]
            t2new_tmp -= einsum('akic,kjbc->ijab', Wvoov[ka, kk, ki], t2[kk, kj, kb])

            kc = kconserv[kk, ka, kj]
            t2new_tmp -= einsum('bkci,kjac->ijab', Wvovo[kb, kk, kc], t2[kk, kj, ka])
        return t2new_tmp

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ki, kj, ka, kb = kq
        t2new_tmp = _t2_voov2(ki,kj,ka,kb)
        t2new[ki, kj, ka] += t2new_tmp
        if lib.isin_1d((kj,ki,kb,ka), kqrts.kqrts_ibz):
            t2new[kj, ki, kb] += t2new_tmp.transpose(1, 0, 3, 2)
        else:
            t2new_tmp = _t2_voov2(kj,ki,kb,ka)
            t2new[ki, kj, ka] += t2new_tmp.transpose(1, 0, 3, 2)

    Wvoov = Wvovo = None

    for i, kq in enumerate(kqrts.kqrts_ibz):
        ki, kj, ka, kb = kq
        eia = _get_epq([0,nocc,ki,mo_e_o,nonzero_opadding],
                       [0,nvir,ka,mo_e_v,nonzero_vpadding],
                       fac=[1.0,-1.0])
        ejb = _get_epq([0,nocc,kj,mo_e_o,nonzero_opadding],
                       [0,nvir,kb,mo_e_v,nonzero_vpadding],
                       fac=[1.0,-1.0])
        eijab = eia[:, None, :, None] + ejb[:, None, :]
        t2new[ki, kj, ka] /= eijab

    return t1new, t2new

def add_vvvv_(cc, Ht2, t1, t2, eris):
    kpts = cc.kpts
    kqrts = cc.kqrts
    rmat = cc.rmat

    nocc = cc.nocc
    nmo = cc.nmo
    nvir = nmo - nocc
    nkpts = kpts.nkpts
    kconserv = cc.khelper.kconserv

    mem_now = lib.current_memory()[0]
    if (not cc.incore_complete and
        cc.direct and getattr(eris, 'Lpv', None) is not None):
        def get_Wvvvv(ka, kb, kc):
            Lpv = eris.Lpv
            kd = kconserv[ka, kc, kb]
            Lbd = (Lpv[kb,kd][:,nocc:] -
                   lib.einsum('Lkd,kb->Lbd', Lpv[kb,kd][:,:nocc], t1[kb]))
            Wvvvv = lib.einsum('Lac,Lbd->abcd', Lpv[ka,kc][:,nocc:], Lbd)
            Lbd = None
            kcbd = lib.einsum('Lkc,Lbd->kcbd', Lpv[ka,kc][:,:nocc],
                              Lpv[kb,kd][:,nocc:])
            Wvvvv -= lib.einsum('kcbd,ka->abcd', kcbd, t1[ka])
            Wvvvv *= (1. / nkpts)
            return Wvvvv
    elif (cc.incore_complete or
          _memory_4d(cc, [nvir,]*4) + mem_now < cc.max_memory * .9):
        _Wvvvv = imdk.cc_Wvvvv(kpts, kqrts, t1, t2, eris, rmat)

        mem_now = lib.current_memory()[0]
        if (not cc.ktensor_direct and
            _memory_4d(cc, [nvir,]*4, False) + mem_now < cc.max_memory * .9):
            _Wvvvv = _Wvvvv.todense()

        get_Wvvvv = lambda ka, kb, kc: _Wvvvv[ka, kb, kc]
    else:
        metadata = {'kpts': kpts, 'kqrts': kqrts, 'rmat': rmat,
                    'label': 'vvvv', 'trans': 'ccnn',
                    'incore': False}
        _Wvvvv = ktensor.empty([nvir,]*4, dtype=t1.dtype, metadata=metadata)
        _Wvvvv = imdk.cc_Wvvvv(kpts, kqrts, t1, t2, eris, rmat, _Wvvvv)

        mem_now = lib.current_memory()[0]
        if (not cc.ktensor_direct and
            _memory_4d(cc, [nvir,]*4, False) + mem_now < cc.max_memory * .9):
            _Wvvvv = _Wvvvv.todense()

        get_Wvvvv = lambda ka, kb, kc: _Wvvvv[ka, kb, kc]

    kakb, igroup = np.unique(kqrts.kqrts_ibz[:,2:], axis=0, return_inverse=True)
    igroup = igroup.ravel()
    for i in range(np.amax(igroup) + 1):
        ka, kb = kakb[i]
        idx = np.where(igroup==i)[0]

        for kc in range(nkpts):
            kd = kconserv[ka, kc, kb]
            Wvvvv = get_Wvvvv(ka, kb, kc)
            for m in idx:
                ki,kj,kaa,kbb = kqrts.kqrts_ibz[m]
                assert kaa==ka and kbb==kb
                tau = t2[ki, kj, kc].copy()
                if ki == kc and kj == kd:
                    tau += np.einsum('ic,jd->ijcd', t1[ki], t1[kj])
                Ht2[ki, kj, ka] += einsum('abcd,ijcd->ijab', Wvvvv, tau)

    _Wvvvv = None
    return Ht2

def energy(cc, t1, t2, eris):
    kpts = cc.kpts
    kqrts = cc.kqrts

    nkpts, nocc, nvir = t1.shape
    fock = eris.fock
    e = 0

    for ki_ibz in range(kpts.nkpts_ibz):
        ki = kpts.ibz2bz[ki_ibz]
        weight = kpts.weights_ibz[ki_ibz]
        e += 2 * einsum('ia,ia', fock[ki,:nocc,nocc:], t1[ki]) * weight

    tau = ktensor.empty_like(t2)
    for kq in kqrts.kqrts_ibz:
        ki, kj, ka, kb = kq
        tau[ki, kj, ka] = t2[ki, kj, ka]
        if ki == ka and kj == kb:
            tau[ki, kj, ka] += einsum('ia,jb->ijab', t1[ki], t1[kj])

    kq_weights = kqrts.weights_ibz
    for k, kq in enumerate(kqrts.kqrts_ibz):
        ki, kj, ka, kb = kq
        weight = kq_weights[k] * nkpts**3
        e += 2 * einsum('ijab,ijab', tau[ki, kj, ka], eris.oovv[ki, kj, ka]) * weight
        e += -einsum('ijab,ijba', tau[ki, kj, ka], eris.oovv[ki, kj, kb]) * weight

    e /= nkpts
    if abs(e.imag) > 1e-4:
        logger.warn(cc, 'Non-zero imaginary part found in KRCCSD energy %s', e)
    return e.real


class KsymAdaptedRCCSD(RCCSD):
    _keys = {'kqrts', 'rmat', 'ktensor_direct', 'eris_outcore'}

    def __init__(self, mf, frozen=None, mo_coeff=None, mo_occ=None):
        '''
        Attributes:
            ktensor_direct : bool
                If set to True, the tensors will be stored as block-sparse,
                and the symmetry related blocks are computed on-the-fly when needed.
                Otherwise, the tensors will be converted to dense tensors whenever
                there is enough memory. Default is False.
            eris_outcore : bool
                If set to True, the integrals will be always stored on the disk.
                Otherwise, whether the integrals are stored on the disk or in memory
                depends on the available memory size. Default is False.
        '''
        # NOTE self._scf is a non-symmetry object, see RCCSD.__init__
        RCCSD.__init__(self, mf, frozen, mo_coeff, mo_occ)
        self.kqrts = KQuartets(mf.kpts).build()
        self.rmat = None
        self.ktensor_direct = False
        self.eris_outcore = False

    def ao2mo(self, mo_coeff=None):
        eris = _PhysicistsERIs()
        eris._common_init_(self, mo_coeff)

        # use padded mo_coeff to construct the rotation matrix
        self.rmat = MORotationMatrix(self.kpts, eris.mo_coeff,
                                     self._scf.get_ovlp(), eris.nocc, eris.nmo)
        self.rmat.build()

        mem_now = lib.current_memory()[0]
        mem_incore = _mem_usage(self)
        if not self.eris_outcore and (
                self.incore_complete or mem_incore + mem_now < self.max_memory):
            eris = _make_eris_incore(self, eris, self._scf.with_df.ao2mo)
        else:
            eris = _make_eris_outcore(self, eris, self._scf.with_df.ao2mo)
        return eris

    def init_amps(self, eris):
        time0 = logger.process_clock(), logger.perf_counter()
        nocc = self.nocc
        nvir = self.nmo - nocc
        nkpts = self.nkpts

        kpts = self.kpts
        kqrts = self.kqrts
        rmat = self.rmat
        assert rmat is not None

        metadata = {'kpts': kpts, 'rmat': rmat,
                    'label': 'ov', 'trans': 'nc', 'incore': True}
        t1 = ktensor.zeros((nocc, nvir), dtype=eris.fock.dtype,
                           metadata=metadata)

        metadata = {'kpts': kpts, 'kqrts': kqrts, 'rmat': rmat,
                    'label': 'oovv', 'trans': 'nncc', 'incore': True}
        t2 = ktensor.empty((nocc,nocc,nvir,nvir), dtype=eris.fock.dtype,
                           metadata=metadata)
        mo_e_o = [eris.mo_energy[k][:nocc] for k in range(nkpts)]
        mo_e_v = [eris.mo_energy[k][nocc:] for k in range(nkpts)]

        # Get location of padded elements in occupied and virtual space
        nonzero_opadding, nonzero_vpadding = padding_k_idx(self, kind="split")

        emp2 = 0
        for i, kq in enumerate(kqrts.kqrts_ibz):
            ki, kj, ka, kb = kq
            weight = kqrts.weights_ibz[i] * nkpts**3

            eia = _get_epq([0,nocc,ki,mo_e_o,nonzero_opadding],
                           [0,nvir,ka,mo_e_v,nonzero_vpadding],
                           fac=[1.0,-1.0])

            ejb = _get_epq([0,nocc,kj,mo_e_o,nonzero_opadding],
                           [0,nvir,kb,mo_e_v,nonzero_vpadding],
                           fac=[1.0,-1.0])
            eijab = eia[:, None, :, None] + ejb[:, None, :]

            eris_ijab = eris.oovv[ki, kj, ka]
            eris_ijba = eris.oovv[ki, kj, kb]
            t2[ki, kj, ka] = eris_ijab.conj() / eijab
            woovv = 2 * eris_ijab - eris_ijba.transpose(0, 1, 3, 2)
            emp2 += np.einsum('ijab,ijab', t2[ki, kj, ka], woovv) * weight

        self.emp2 = emp2.real / nkpts
        logger.info(self, 'Init t2, MP2 energy (with fock eigenvalue shift) = %.15g', self.emp2)
        logger.timer(self, 'init mp2', *time0)
        return self.emp2, t1, t2

    def amplitudes_to_vector(self, t1, t2):
        t1_raw = np.asarray(getattr(t1, 'data', t1))
        t2_raw = np.asarray(getattr(t2, 'data', t2))
        return np.concatenate((t1_raw, t2_raw), axis=None)

    def vector_to_amplitudes(self, vec):
        kpts = self.kpts
        kqrts = self.kqrts
        rmat = self.rmat
        nocc = self.nocc
        nvir = self.nmo - nocc
        t1_size = kpts.nkpts_ibz * nocc * nvir
        t1_flat = vec[:t1_size]
        t2_flat = vec[t1_size:]

        metadata = {'kpts': kpts, 'rmat': rmat,
                    'label': 'ov', 'trans': 'nc', 'incore': True}
        t1 = ktensor.fromraw(t1_flat, (nocc,nvir), dtype=vec.dtype,
                             metadata=metadata)

        metadata = {'kpts': kpts, 'kqrts': kqrts, 'rmat': rmat,
                    'label': 'oovv', 'trans': 'nncc', 'incore': True}
        t2 = ktensor.fromraw(t2_flat, (nocc,nocc,nvir,nvir), dtype=vec.dtype,
                             metadata=metadata)
        return t1, t2

    energy = energy
    update_amps = update_amps


def _make_eris_incore(cc, eris, fao2mo):
    log = logger.Logger(cc.stdout, cc.verbose)
    log.info('using incore ERI storage')
    cput0 = (logger.process_clock(), logger.perf_counter())

    mo_coeff = eris.mo_coeff
    nocc = eris.nocc
    nvir = eris.nvir
    nmo = eris.nmo
    dtype = eris.dtype

    common_metadata = {'kpts'  : cc.kpts,
                       'kqrts' : cc.kqrts,
                       'rmat'  : cc.rmat,
                       'trans' : 'ccnn',
                       'incore': True}

    eris.oooo = ktensor.empty([nocc,nocc,nocc,nocc], dtype=dtype,
                              metadata={**common_metadata, 'label': 'oooo'})
    eris.ooov = ktensor.empty([nocc,nocc,nocc,nvir], dtype=dtype,
                              metadata={**common_metadata, 'label': 'ooov'})
    eris.oovv = ktensor.empty([nocc,nocc,nvir,nvir], dtype=dtype,
                              metadata={**common_metadata, 'label': 'oovv'})
    eris.ovov = ktensor.empty([nocc,nvir,nocc,nvir], dtype=dtype,
                              metadata={**common_metadata, 'label': 'ovov'})
    eris.voov = ktensor.empty([nvir,nocc,nocc,nvir], dtype=dtype,
                              metadata={**common_metadata, 'label': 'voov'})
    eris.vovv = ktensor.empty([nvir,nocc,nvir,nvir], dtype=dtype,
                              metadata={**common_metadata, 'label': 'vovv'})
    eris.vvvv = ktensor.empty([nvir,nvir,nvir,nvir], dtype=dtype,
                              metadata={**common_metadata, 'label': 'vvvv'})

    kpts = cc.kpts.kpts
    nkpts = len(kpts)
    kqrts = cc.kqrts
    khelper = cc.khelper
    kconserv = khelper.kconserv
    kptlist = kqrts.kqrts_ibz[:,:3][:,[0,2,1]] #chemists' notation
    khelper.build_symm_map(kptlist=kptlist)

    for (iki, ika, ikj) in khelper.symm_map.keys():
        ikb = kconserv[iki, ika, ikj]
        eri_kpt = fao2mo((mo_coeff[iki],mo_coeff[ika],mo_coeff[ikj],mo_coeff[ikb]),
                         (kpts[iki],kpts[ika],kpts[ikj],kpts[ikb]), compact=False)

        if np.issubdtype(dtype, np.floating):
            eri_kpt = eri_kpt.real
        eri_kpt = eri_kpt.reshape(nmo, nmo, nmo, nmo)

        for (ki, ka, kj) in khelper.symm_map[(iki, ika, ikj)]:
            eri_kpt_symm = khelper.transform_symm(eri_kpt, ki, ka, kj).transpose(0, 2, 1, 3)
            eris.oooo[ki, kj, ka] = eri_kpt_symm[:nocc, :nocc, :nocc, :nocc] / nkpts
            eris.ooov[ki, kj, ka] = eri_kpt_symm[:nocc, :nocc, :nocc, nocc:] / nkpts
            eris.oovv[ki, kj, ka] = eri_kpt_symm[:nocc, :nocc, nocc:, nocc:] / nkpts
            eris.ovov[ki, kj, ka] = eri_kpt_symm[:nocc, nocc:, :nocc, nocc:] / nkpts
            eris.voov[ki, kj, ka] = eri_kpt_symm[nocc:, :nocc, :nocc, nocc:] / nkpts
            eris.vovv[ki, kj, ka] = eri_kpt_symm[nocc:, :nocc, nocc:, nocc:] / nkpts
            eris.vvvv[ki, kj, ka] = eri_kpt_symm[nocc:, nocc:, nocc:, nocc:] / nkpts

    if not cc.ktensor_direct:
        eris.oooo = eris.oooo.todense()
        eris.ooov = eris.ooov.todense()
        eris.oovv = eris.oovv.todense()
        eris.ovov = eris.ovov.todense()
        eris.voov = eris.voov.todense()
        eris.vovv = eris.vovv.todense()
        eris.vvvv = eris.vvvv.todense()

    log.timer('CCSD integral transformation', *cput0)
    return eris

def _make_eris_outcore(cc, eris, fao2mo):
    log = logger.Logger(cc.stdout, cc.verbose)
    log.info('using outcore ERI storage')
    cput0 = (logger.process_clock(), logger.perf_counter())

    mo_coeff = eris.mo_coeff
    nocc = eris.nocc
    nvir = eris.nvir
    nmo = eris.nmo
    dtype = eris.dtype

    common_metadata = {'kpts'  : cc.kpts,
                       'kqrts' : cc.kqrts,
                       'rmat'  : cc.rmat,
                       'trans' : 'ccnn',
                       'incore': False}

    eris.oooo = ktensor.empty([nocc,nocc,nocc,nocc], dtype=dtype,
                              metadata={**common_metadata, 'label': 'oooo'})
    eris.ooov = ktensor.empty([nocc,nocc,nocc,nvir], dtype=dtype,
                              metadata={**common_metadata, 'label': 'ooov'})
    eris.oovv = ktensor.empty([nocc,nocc,nvir,nvir], dtype=dtype,
                              metadata={**common_metadata, 'label': 'oovv'})
    eris.ovov = ktensor.empty([nocc,nvir,nocc,nvir], dtype=dtype,
                              metadata={**common_metadata, 'label': 'ovov'})
    eris.voov = ktensor.empty([nvir,nocc,nocc,nvir], dtype=dtype,
                              metadata={**common_metadata, 'label': 'voov'})
    eris.vovv = ktensor.empty([nvir,nocc,nvir,nvir], dtype=dtype,
                              metadata={**common_metadata, 'label': 'vovv'})

    vvvv_required = (not cc.direct
                     or not isinstance(cc._scf.with_df, (GDF, RSGDF))
                     or cc._scf.cell.dimension == 2)
    if vvvv_required:
        eris.vvvv = ktensor.empty([nvir,nvir,nvir,nvir], dtype=dtype,
                                  metadata={**common_metadata, 'label': 'vvvv'})

    kpts = cc.kpts.kpts
    nkpts = len(kpts)
    kqrts = cc.kqrts

    # <ij|pq>  = (ip|jq)
    cput1 = logger.process_clock(), logger.perf_counter()
    for kprqs in kqrts.kqrts_ibz:
        kp, kr, kq, ks = kprqs
        orbo_p = mo_coeff[kp][:, :nocc]
        orbo_r = mo_coeff[kr][:, :nocc]
        buf_kpt = fao2mo((orbo_p, mo_coeff[kq], orbo_r, mo_coeff[ks]),
                         (kpts[kp], kpts[kq], kpts[kr], kpts[ks]), compact=False)
        if np.issubdtype(dtype, np.floating):
            buf_kpt = buf_kpt.real
        buf_kpt = buf_kpt.reshape(nocc, nmo, nocc, nmo).transpose(0, 2, 1, 3)
        eris.oooo[kp, kr, kq] = buf_kpt[:, :, :nocc, :nocc] / nkpts
        eris.ooov[kp, kr, kq] = buf_kpt[:, :, :nocc, nocc:] / nkpts
        eris.oovv[kp, kr, kq] = buf_kpt[:, :, nocc:, nocc:] / nkpts
    cput1 = log.timer_debug1('transforming oopq', *cput1)

    # <ia|jb> = (ij|ab)
    for kiajb in kqrts.kqrts_ibz:
        ki, ka, kj, kb = kiajb
        orb_i = mo_coeff[ki][:, :nocc]
        orb_a = mo_coeff[ka][:, nocc:]
        orb_j = mo_coeff[kj][:, :nocc]
        orb_b = mo_coeff[kb][:, nocc:]
        buf_kpt = fao2mo((orb_i, orb_j, orb_a, orb_b),
                         (kpts[ki], kpts[kj], kpts[ka], kpts[kb]), compact=False)
        if np.issubdtype(dtype, np.floating):
            buf_kpt = buf_kpt.real
        buf_kpt = buf_kpt.reshape(nocc, nocc, nvir, nvir).transpose(0, 2, 1, 3)
        eris.ovov[ki, ka, kj] = buf_kpt / nkpts
    cput1 = log.timer_debug1('transforming ovov', *cput1)

    # <ai|pq> = (ap|iq)
    for kaipq in kqrts.kqrts_ibz:
        ka, ki, kp, kq = kaipq
        orb_a = mo_coeff[ka][:, nocc:]
        orb_i = mo_coeff[ki][:, :nocc]
        buf_kpt = fao2mo((orb_a, mo_coeff[kp], orb_i, mo_coeff[kq]),
                         (kpts[ka], kpts[kp], kpts[ki], kpts[kq]), compact=False)
        if np.issubdtype(dtype, np.floating):
            buf_kpt = buf_kpt.real
        buf_kpt = buf_kpt.reshape(nvir, nmo, nocc, nmo).transpose(0, 2, 1, 3)
        # TODO: compute vovv on the fly
        eris.vovv[ka, ki, kp] = buf_kpt[:, :, nocc:, nocc:] / nkpts
        eris.voov[ka, ki, kp] = buf_kpt[:, :, :nocc, nocc:] / nkpts
    cput1 = log.timer_debug1('transforming vopq', *cput1)

    mem_now = lib.current_memory()[0]
    if not vvvv_required:
        _init_df_eris(cc, eris)
    elif nvir ** 4 * 16 / 1e6 + mem_now < cc.max_memory:
        khelper = cc.khelper
        kconserv = khelper.kconserv
        kptlist = kqrts.kqrts_ibz[:,:3][:,[0,2,1]] #chemists' notation
        khelper.build_symm_map(kptlist=kptlist)
        for (ikp, ikq, ikr) in khelper.symm_map.keys():
            iks = kconserv[ikp, ikq, ikr]
            orbv_p = mo_coeff[ikp][:, nocc:]
            orbv_q = mo_coeff[ikq][:, nocc:]
            orbv_r = mo_coeff[ikr][:, nocc:]
            orbv_s = mo_coeff[iks][:, nocc:]
            buf_kpt = fao2mo((orbv_p,orbv_q,orbv_r,orbv_s),
                             kpts[[ikp,ikq,ikr,iks]], compact=False)
            if np.issubdtype(dtype, np.floating):
                buf_kpt = buf_kpt.real
            buf_kpt = buf_kpt.reshape([nvir,]*4)
            for (kp, kq, kr) in khelper.symm_map[(ikp, ikq, ikr)]:
                buf_kpt_symm = khelper.transform_symm(buf_kpt, kp, kq, kr).transpose(0, 2, 1, 3)
                eris.vvvv[kp, kr, kq] = buf_kpt_symm / nkpts
    else:
        raise MemoryError(f'Minimal memory requirements '
                          f'{mem_now + nvir ** 4 / 1e6 * 16 * 2} MB')
    cput1 = log.timer_debug1('transforming vvvv', *cput1)

    log.timer('CCSD integral transformation', *cput0)
    return eris


class _PhysicistsERIs():
    def __init__(self, cell=None):
        self.cell = cell
        self.kpts = None
        self.mo_coeff = None
        self.nocc = None
        self.nmo = None
        self.nvir = None
        self.fock = None
        self.dtype = None

        self.oooo = None
        self.ooov = None
        self.oovv = None
        self.ovov = None
        self.voov = None
        self.vovv = None
        self.vvvv = None
        self.Lpv = None

    def _common_init_(self, cc, mo_coeff=None):
        from pyscf.pbc import tools
        from pyscf.pbc.cc.ccsd import _adjust_occ
        mf = cc._scf
        self.cell = cell = mf.cell
        self.kpts = kpts = cc.kpts
        self.nocc = nocc = cc.nocc
        self.nmo = nmo = cc.nmo
        self.nvir = nmo - nocc

        if mo_coeff is None:
            mo_coeff = cc.mo_coeff
        self.dtype = mo_coeff[-1].dtype

        # Re-make our fock MO matrix elements from density and fock AO
        # FIXME what if mo_coeff is not consistent with cc.mo_occ?
        dm = mf.make_rdm1(mo_coeff, cc.mo_occ)
        exxdiv = mf.exxdiv if cc.keep_exxdiv else None
        with lib.temporary_env(mf, exxdiv=exxdiv):
            # _scf.exxdiv affects eris.fock. HF exchange correction should be
            # excluded from the Fock matrix.
            vhf = mf.get_veff(cell, dm)
        fockao = mf.get_hcore() + vhf

        self.mo_coeff = mo_coeff = padded_mo_coeff(cc, mo_coeff)

        fock = np.asarray([reduce(np.dot, (mo.T.conj(), fockao[k], mo))
                          for k, mo in enumerate(mo_coeff)])
        self.fock = fock

        mo_energy = [fock[k].diagonal().real for k in range(len(fock))]
        if not cc.keep_exxdiv:
            # Add HFX correction in the self.mo_energy to improve convergence in
            # CCSD iteration. It is useful for the 2D systems since their occupied and
            # the virtual orbital energies may overlap which may lead to numerical
            # issue in the CCSD iterations.
            # FIXME: Whether to add this correction for other exxdiv treatments?
            # Without the correction, MP2 energy may be largely off the correct value.
            madelung = tools.madelung(cell, kpts.kpts)
            mo_energy = [_adjust_occ(mo_e, nocc, -madelung)
                         for k, mo_e in enumerate(mo_energy)]

        # Get location of padded elements in occupied and virtual space.
        nocc_per_kpt = get_nocc(cc, per_kpoint=True)
        nonzero_padding = padding_k_idx(cc, kind="joint")

        # Check direct and indirect gaps for possible issues with CCSD convergence.
        mo_e = [mo_energy[kp][nonzero_padding[kp]] for kp in range(len(mo_energy))]
        mo_e = np.sort([y for x in mo_e for y in x])  # Sort de-nested array
        gap = mo_e[np.sum(nocc_per_kpt)] - mo_e[np.sum(nocc_per_kpt)-1]
        if gap < 1e-5:
            logger.warn(cc, 'HOMO-LUMO gap %s too small for KCCSD. '
                            'May cause issues in convergence.', gap)
        self.mo_energy = mo_energy


def _mem_usage(cc):
    nocc = cc.nocc
    nmo = cc.nmo
    nvir = nmo - nocc
    nkpts = cc.kpts.nkpts
    #Wvvvv, eris.vvvv
    incore  = _memory_4d(cc, [nvir,]*4) * 2
    #eris.vovv
    incore += _memory_4d(cc, [nvir,nvir,nvir,nocc])
    #Wvoov, Wvovo, eris.oovv, eris.ovov, eris.voov, t2new, t2
    incore += _memory_4d(cc, [nvir,nvir,nocc,nocc]) * 7
    #eris.ooov
    incore += _memory_4d(cc, [nvir,nocc,nocc,nocc])
    #Woooo, eris.oooo
    incore += _memory_4d(cc, [nocc,]*4) * 2

    if not cc.ktensor_direct:
        incore += _memory_4d(cc, [nvir,]*4, False) * 2
        incore += _memory_4d(cc, [nvir,nvir,nvir,nocc], False)
        incore += _memory_4d(cc, [nvir,nvir,nocc,nocc], False) * 5
        incore += _memory_4d(cc, [nvir,nocc,nocc,nocc], False)
        incore += _memory_4d(cc, [nocc,]*4, False) * 2

    #temp
    incore += nkpts * nmo**4 * 16 / 1e6
    #1e, fock, t1
    incore += nkpts * nmo**2 * 16 / 1e6 * 7
    #t2_bz
    incore += _memory_4d(cc, [nvir,nvir,nocc,nocc], False)

    logger.info(cc, f"Incore memory estimation: {incore} MB")
    return incore

def _memory_4d(cc, shape, ibz=True):
    if ibz:
        nk = len(cc.kqrts.kqrts_ibz)
    else:
        nk = cc.kpts.nkpts**3
    return nk * np.prod(shape) * 16 / 1e6



#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from pyscf.pbc import scf, dft, gto
from pyscf.eph.rhf import solve_hmat, _freq_mass_weighted_vec
from pyscf.lib import logger, param
from pyscf.data.nist import MP_ME

'''Electron-Phonon matrix from finite difference for Gamma Point'''
# Note, the code now only return eph matrix at Gamma Point
# cell relaxation needs to be performed before computing eph matrix

def copy_mf(mf, cell):
    mf1 = mf.__class__(cell)
    if isinstance(mf, scf.hf.KohnShamDFT):
        mf1.xc = mf.xc
    mf1.kpts = mf.kpts
    mf1.exxdiv = getattr(mf, 'exxdiv', None)
    mf1.conv_tol = mf.conv_tol
    mf1.conv_tol_grad = mf.conv_tol_grad
    return mf1

def run_mfs(mf, cells_a, cells_b):
    '''perform a set of calculations on given two sets of cell'''
    nconfigs = len(cells_a)
    dm0 = mf.make_rdm1()
    mflist = []
    for i in range(nconfigs):
        mf1 = copy_mf(mf, cells_a[i])
        mf2 = copy_mf(mf, cells_b[i])
        mf1.kernel(dm0=dm0)
        mf2.kernel(dm0=dm0)
        if not (mf1.converged):
            logger.warn(mf, "%ith config mf1 not converged", i)
        if not (mf2.converged):
            logger.warn(mf, "%ith config mf2 not converged", i)
        mflist.append((mf1, mf2))
    return mflist

def gen_cells(cell, disp):
    """From the given cell, generate 3N cells with a shift on
    + displacement(cell_a) and - displacement(cell_s) on each Cartesian
    coordinates
    """
    coords = cell.atom_coords()
    if cell.unit[0].lower() == 'a':
        coords = np.asarray(coords) * param.BOHR
        disp_ = disp * param.BOHR
    else:
        disp_ = disp
    natoms = len(coords)
    cell_a, cell_s = [],[]
    for i in range(natoms):
        for x in range(3):
            new_coords_a, new_coords_s = coords.copy(), coords.copy()
            new_coords_a[i][x] += disp_
            new_coords_s[i][x] -= disp_
            atoma = [[cell.atom_symbol(j), coord] for (j, coord) in zip(range(natoms), new_coords_a)]
            atoms = [[cell.atom_symbol(j), coord] for (j, coord) in zip(range(natoms), new_coords_s)]
            cell_a.append(cell.set_geom_(atoma, inplace=False))
            cell_s.append(cell.set_geom_(atoms, inplace=False))
    return cell_a, cell_s

def get_vmat(mf, mfset, disp):
    nconfigs = len(mfset)
    vmat=[]
    mygrad = mf.nuc_grad_method()
    veff  = mygrad.get_veff()
    RESTRICTED = (veff.ndim==4)
    v1e = mygrad.get_hcore() - np.asarray(mf.cell.pbc_intor("int1e_ipkin", kpts=mf.kpts))
    if RESTRICTED:
        vtmp = veff - v1e.transpose(1,0,2,3)
    else:
        vtmp = veff - v1e.transpose(1,0,2,3)[:,None]

    aoslice = mf.cell.aoslice_by_atom()
    for i in range(nconfigs):
        atmid, axis = np.divmod(i, 3)
        p0, p1 = aoslice[atmid][2:]
        mf1, mf2 = mfset[i]
        if RESTRICTED:
            vfull1 = mf1.get_veff() + mf1.get_hcore() \
                   - np.asarray(mf1.cell.pbc_intor('int1e_kin', kpts=mf1.kpts))  # <u+|V+|v+>
            vfull2 = mf2.get_veff() + mf2.get_hcore() \
                   - np.asarray(mf2.cell.pbc_intor('int1e_kin', kpts=mf2.kpts))  # <u-|V-|v->
        else:
            vfull1 = mf1.get_veff() + mf1.get_hcore()[None] \
                   - np.asarray(mf1.cell.pbc_intor('int1e_kin', kpts=mf1.kpts))[None]  # <u+|V+|v+>
            vfull2 = mf2.get_veff() + mf2.get_hcore()[None] \
                   - np.asarray(mf2.cell.pbc_intor('int1e_kin', kpts=mf2.kpts))[None]  # <u-|V-|v->
        vfull = (vfull1 - vfull2)/disp  # (<p+|V+|q+>-<p-|V-|q->)/dR
        if RESTRICTED:
            vfull[:,p0:p1] -= vtmp[axis,:,p0:p1]
            vfull[:,:,p0:p1] -= vtmp[axis,:,p0:p1].transpose(0,2,1).conj()
        else:
            vfull[:,:,p0:p1] -= vtmp[axis,:,:,p0:p1]
            vfull[:,:,:,p0:p1] -= vtmp[axis,:,:,p0:p1].transpose(0,1,3,2).conj()

        vmat.append(vfull)

    vmat= np.asarray(vmat)
    if RESTRICTED:
        return vmat[:,0]
    else:
        return vmat[:,:,0]

def run_hess(mfset, disp):
    natoms = len(mfset[0][0].cell.atom_mass_list())
    hess=[]
    for (mf1, mf2) in mfset:
        grad1 = mf1.nuc_grad_method()
        grad2 = mf2.nuc_grad_method()
        g1 = grad1.kernel()
        g2 = grad2.kernel()
        gdelta = (g1-g2) / disp
        hess.append(gdelta)
    hess = np.asarray(hess).reshape(natoms, 3, natoms, 3).transpose(0,2,1,3)
    return hess


def kernel(mf, disp=1e-4, mo_rep=False):
    if not mf.converged: mf.kernel()
    mo_coeff = np.asarray(mf.mo_coeff)
    RESTRICTED= (mo_coeff.ndim==3)
    cell = mf.cell
    cells_a, cells_b = gen_cells(cell, disp/2.0) # generate a bunch of cells with disp/2 on each cartesian coord
    mfset = run_mfs(mf, cells_a, cells_b) # run mean field calculations on all these cells
    vmat = get_vmat(mf, mfset, disp) # extracting <u|dV|v>/dR
    hmat = run_hess(mfset, disp)
    omega, vec = solve_hmat(cell, hmat)
    mass = cell.atom_mass_list() * MP_ME
    vec = _freq_mass_weighted_vec(vec, omega, mass)
    if mo_rep:
        if RESTRICTED:
            vmat = np.einsum('xuv,up,vq->xpq', vmat, mo_coeff[0].conj(), mo_coeff[0])
        else:
            vmat = np.einsum('xsuv,sup,svq->xspq', vmat, mo_coeff[:,0].conj(), mo_coeff[:,0])

    if RESTRICTED:
        mat = np.einsum('xJ,xpq->Jpq', vec, vmat)
    else:
        mat = np.einsum('xJ,xspq->sJpq', vec, vmat)
    return mat, omega

if __name__ == '__main__':
    cell = gto.Cell()
    cell.atom='''
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    '''
    cell.basis = 'gth-szv'
    cell.pseudo = 'gth-pade'
    cell.a = '''
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000'''
    cell.unit = 'B'
    cell.verbose = 4
    cell.build()

    kpts = cell.make_kpts([1,1,1])
    mf = dft.KRKS(cell, kpts)
    mf.conv_tol = 1e-14
    mf.conv_tol_grad = 1e-8
    mf.kernel()

    vmat, omega = kernel(mf, mo_rep=True)



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

from functools import reduce
import numpy
import scipy.linalg
from pyscf.lib import param
from pyscf.lib import logger
from pyscf import gto
from pyscf import __config__

REF_BASIS = getattr(__config__, 'lo_orth_pre_orth_ao_method', 'ANO')
ORTH_METHOD = getattr(__config__, 'lo_orth_orth_ao_method', 'meta_lowdin')
PROJECT_ECP_BASIS = getattr(__config__, 'lo_orth_project_ecp_basis', True)


def lowdin(s):
    ''' new basis is |mu> c^{lowdin}_{mu i} '''
    e, v = scipy.linalg.eigh(s)
    idx = e > 1e-15
    return numpy.dot(v[:,idx]/numpy.sqrt(e[idx]), v[:,idx].conj().T)

def schmidt(s):
    c = numpy.linalg.cholesky(s)
    return scipy.linalg.solve_triangular(c, numpy.eye(c.shape[1]), lower=True,
                                         overwrite_b=False).conj().T

def vec_lowdin(c, s=1):
    ''' lowdin orth for the metric c.T*s*c and get x, then c*x'''
    #u, w, vh = numpy.linalg.svd(c)
    #return numpy.dot(u, vh)
    # svd is slower than eigh
    return numpy.dot(c, lowdin(reduce(numpy.dot, (c.conj().T,s,c))))

def vec_schmidt(c, s=1):
    ''' schmidt orth for the metric c.T*s*c and get x, then c*x'''
    if isinstance(s, numpy.ndarray):
        return numpy.dot(c, schmidt(reduce(numpy.dot, (c.conj().T,s,c))))
    else:
        return numpy.linalg.qr(c)[0]

def weight_orth(s, weight):
    ''' new basis is |mu> c_{mu i}, c = w[(wsw)^{-1/2}]'''
    s1 = weight[:,None] * s * weight
    c = lowdin(s1)
    return weight[:,None] * c


def pre_orth_ao(mol, method=REF_BASIS):
    '''Restore AO characters.  Possible methods include the ANO/MINAO
    projection or fraction-averaged atomic RHF calculation'''
    if isinstance(method, str) and method.upper() == 'SCF':
        return pre_orth_ao_atm_scf(mol)
    else:
        # Use ANO/MINAO basis to define the strongly occupied set
        return project_to_atomic_orbitals(mol, method)
restore_ao_character = pre_orth_ao

def project_to_atomic_orbitals(mol, ref_basis):
    '''projected AO = |bas><bas|ANO>

    args:
        ref_basis : str or basis dict
            Name, or filename, or a dict of reference basis set
    '''
    from pyscf.scf.addons import project_mo_nr2nr
    from pyscf.scf import atom_hf
    from pyscf.gto.ecp import core_configuration

    def search_atm_l(atm, l):
        bas_ang = atm._bas[:,gto.ANG_OF]
        ao_loc = atm.ao_loc_nr()
        idx = []
        for ib in numpy.where(bas_ang == l)[0]:
            idx.extend(range(ao_loc[ib], ao_loc[ib+1]))
        return idx

    # Overlap of ANO and ECP basis
    def ecp_ano_det_ovlp(atm_ecp, atm_ano, ecpcore):
        ecp_ao_loc = atm_ecp.ao_loc_nr()
        ano_ao_loc = atm_ano.ao_loc_nr()
        ecp_ao_dim = ecp_ao_loc[1:] - ecp_ao_loc[:-1]
        ano_ao_dim = ano_ao_loc[1:] - ano_ao_loc[:-1]
        ecp_bas_l = [[atm_ecp.bas_angular(i)]*d for i,d in enumerate(ecp_ao_dim)]
        ano_bas_l = [[atm_ano.bas_angular(i)]*d for i,d in enumerate(ano_ao_dim)]
        ecp_bas_l = numpy.hstack(ecp_bas_l)
        ano_bas_l = numpy.hstack(ano_bas_l)

        nelec_core = 0
        ecp_occ_tmp = []
        ecp_idx = []
        ano_idx = []
        for l in range(4):
            nocc, frac = atom_hf.frac_occ(stdsymb, l)
            l_occ = [2] * ((nocc-ecpcore[l])*(2*l+1))
            if frac > 1e-15:
                l_occ.extend([frac] * (2*l+1))
                nocc += 1
            if nocc == 0:
                break
            nelec_core += 2 * ecpcore[l] * (2*l+1)
            i0 = ecpcore[l] * (2*l+1)
            i1 = nocc * (2*l+1)
            ecp_idx.append(numpy.where(ecp_bas_l==l)[0][:i1-i0])
            ano_idx.append(numpy.where(ano_bas_l==l)[0][i0:i1])
            ecp_occ_tmp.append(l_occ[:i1-i0])
        ecp_idx = numpy.hstack(ecp_idx)
        ano_idx = numpy.hstack(ano_idx)
        ecp_occ = numpy.zeros(atm_ecp.nao_nr())
        ecp_occ[ecp_idx] = numpy.hstack(ecp_occ_tmp)
        nelec_valence_left = int(gto.charge(stdsymb) - nelec_core
                                 - sum(ecp_occ[ecp_idx]))
        if nelec_valence_left > 0:
            logger.warn(mol, 'Characters of %d valence electrons are not identified.\n'
                        'It can affect the "meta-lowdin" localization method '
                        'and the population analysis of SCF method.\n'
                        'Adjustment to the core/valence partition may be needed '
                        '(see function lo.nao.set_atom_conf)\nto get reasonable '
                        'local orbitals or Mulliken population.\n',
                        nelec_valence_left)
            # Return 0 to force the projection to ANO basis
            return 0
        else:
            s12 = gto.intor_cross('int1e_ovlp', atm_ecp, atm_ano)[ecp_idx][:,ano_idx]
            return numpy.linalg.det(s12)

    nelec_ecp_dic = {}
    for ia in range(mol.natm):
        symb = mol.atom_symbol(ia)
        if symb not in nelec_ecp_dic:
            nelec_ecp_dic[symb] = mol.atom_nelec_core(ia)

    basis_converter = gto.mole._generate_basis_converter()
    aos = {}
    atm = gto.Mole()
    atmp = gto.Mole()
    for symb in mol._basis.keys():
        stdsymb = gto.mole._std_symbol(symb)
        atm._atm, atm._bas, atm._env = \
                atm.make_env([[stdsymb,(0,0,0)]], {stdsymb:mol._basis[symb]}, [])
        atm.cart = mol.cart
        atm._built = True
        s0 = atm.intor_symmetric('int1e_ovlp')

        if gto.is_ghost_atom(symb):
            aos[symb] = numpy.diag(1./numpy.sqrt(s0.diagonal()))
            continue

        if isinstance(ref_basis, dict):
            if symb in ref_basis:
                basis_add = basis_converter(stdsymb, ref_basis[symb])
            else:
                basis_add = basis_converter(stdsymb, ref_basis[stdsymb])
        else:
            basis_add = basis_converter(symb, ref_basis)

        atmp._atm, atmp._bas, atmp._env = \
                atmp.make_env([[stdsymb,(0,0,0)]], {stdsymb:basis_add}, [])
        atmp.cart = mol.cart
        atmp._built = True

        if symb in nelec_ecp_dic and nelec_ecp_dic[symb] > 0:
            # If ECP basis has good atomic character, ECP basis can be used in the
            # localization/population analysis directly. Otherwise project ECP
            # basis to ANO basis.
            if not PROJECT_ECP_BASIS:
                continue

            ecpcore = core_configuration(nelec_ecp_dic[symb], atom_symbol=gto.mole._std_symbol(symb))
            # Comparing to ANO valence basis, to check whether the ECP basis set has
            # reasonable AO-character contraction.  The ANO valence AO should have
            # significant overlap to ECP basis if the ECP basis has AO-character.
            if abs(ecp_ano_det_ovlp(atm, atmp, ecpcore)) > .1:
                aos[symb] = numpy.diag(1./numpy.sqrt(s0.diagonal()))
                continue
        else:
            ecpcore = [0] * 4

        # MINAO for heavier elements needs to be used with pseudo potential
        if (ref_basis.upper() == 'MINAO' and
            gto.charge(stdsymb) > 36 and symb not in nelec_ecp_dic):
            raise RuntimeError('Basis MINAO has to be used with ecp for heavy elements')

        ano = project_mo_nr2nr(atmp, numpy.eye(atmp.nao_nr()), atm)
        rm_ano = numpy.eye(ano.shape[0]) - reduce(numpy.dot, (ano, ano.T, s0))
        c = rm_ano.copy()
        for l in range(param.L_MAX):
            idx = numpy.asarray(search_atm_l(atm, l))
            nbf_atm_l = len(idx)
            if nbf_atm_l == 0:
                break

            idxp = numpy.asarray(search_atm_l(atmp, l))
            if l < 4:
                idxp = idxp[ecpcore[l]:]
            nbf_ano_l = len(idxp)

            if mol.cart:
                degen = (l + 1) * (l + 2) // 2
            else:
                degen = l * 2 + 1

            if nbf_atm_l > nbf_ano_l > 0:
                # For angular l, first place the projected ANO, then the rest AOs.
                sdiag = reduce(numpy.dot, (rm_ano[:,idx].T, s0, rm_ano[:,idx])).diagonal()
                nleft = (nbf_atm_l - nbf_ano_l) // degen
                shell_average = numpy.einsum('ij->i', sdiag.reshape(-1,degen))
                shell_rest = numpy.argsort(-shell_average)[:nleft]
                idx_rest = []
                for k in shell_rest:
                    idx_rest.extend(idx[k*degen:(k+1)*degen])
                c[:,idx[:nbf_ano_l]] = ano[:,idxp]
                c[:,idx[nbf_ano_l:]] = rm_ano[:,idx_rest]
            elif nbf_ano_l >= nbf_atm_l > 0:  # More ANOs than the mol basis functions
                c[:,idx] = ano[:,idxp[:nbf_atm_l]]
        sdiag = numpy.einsum('pi,pq,qi->i', c, s0, c)
        c *= 1./numpy.sqrt(sdiag)
        aos[symb] = c

    nao = mol.nao_nr()
    c = numpy.zeros((nao,nao))
    p1 = 0
    for ia in range(mol.natm):
        symb = mol.atom_symbol(ia)
        if symb in mol._basis:
            ano = aos[symb]
        else:
            ano = aos[mol.atom_pure_symbol(ia)]
        p0, p1 = p1, p1 + ano.shape[1]
        c[p0:p1,p0:p1] = ano
    return c
pre_orth_project_ano = project_to_atomic_orbitals

def pre_orth_ao_atm_scf(mol):
    assert (not mol.cart)
    from pyscf.scf import atom_hf
    atm_scf = atom_hf.get_atm_nrhf(mol)
    aoslice = mol.aoslice_by_atom()
    coeff = []
    for ia in range(mol.natm):
        symb = mol.atom_symbol(ia)
        if symb not in atm_scf:
            symb = mol.atom_pure_symbol(ia)

        if symb in atm_scf:
            e_hf, e, c, occ = atm_scf[symb]
        else:  # symb's basis is not specified in the input
            nao_atm = aoslice[ia,3] - aoslice[ia,2]
            c = numpy.zeros((nao_atm, nao_atm))
        coeff.append(c)
    return scipy.linalg.block_diag(*coeff)


def orth_ao(mf_or_mol, method=ORTH_METHOD, pre_orth_ao=REF_BASIS, s=None):
    '''Orthogonalize AOs

    Kwargs:
        method : str
            One of
            | lowdin : Symmetric orthogonalization
            | meta-lowdin : Lowdin orth within core, valence, virtual space separately (JCTC, 10, 3784)
            | NAO

        pre_orth_ao: numpy.ndarray or basis str or basis dict
            Basis functions may not have AO characters. This variable is the
            coefficients to restore AO characters for arbitrary basis. If a
            string of basis name (can be the filename of a basis set) or a
            dict of basis sets are given, they are interpreted as the
            reference basis (by default ANO basis) that the projection
            coefficients are generated based on.  Skip this projection step by
            setting this variable to None.
    '''
    from pyscf import scf
    from pyscf.lo import nao
    if isinstance(mf_or_mol, gto.MoleBase):
        mol = mf_or_mol
        mf = None
    else:
        assert (isinstance(mf_or_mol, scf.hf.SCF))
        mol = mf_or_mol.mol
        mf = mf_or_mol

    if s is None:
        if getattr(mol, 'pbc_intor', None):  # whether mol object is a cell
            s = mol.pbc_intor('int1e_ovlp', hermi=1)
        else:
            s = mol.intor_symmetric('int1e_ovlp')

    if method.lower() == 'lowdin':
        if pre_orth_ao is None:
            c_orth = lowdin(s)
        else:
            if not isinstance(pre_orth_ao, numpy.ndarray):
                pre_orth_ao = restore_ao_character(mol, pre_orth_ao)
            s1 = reduce(numpy.dot, (pre_orth_ao.conj().T, s, pre_orth_ao))
            c_orth = numpy.dot(pre_orth_ao, lowdin(s1))

    elif method.lower() == 'nao':
        assert (mf is not None)
        c_orth = nao.nao(mol, mf, s)

    else:
        # meta_lowdin: partition AOs into core, valence and Rydberg sets,
        # orthogonalizing within each set
        if pre_orth_ao is None:
            pre_orth_ao = numpy.eye(mol.nao)
        elif not isinstance(pre_orth_ao, numpy.ndarray):
            pre_orth_ao = restore_ao_character(mol, pre_orth_ao)
        weight = numpy.ones(pre_orth_ao.shape[0])
        c_orth = nao._nao_sub(mol, weight, pre_orth_ao, s)

    # adjust phase
    for i in range(c_orth.shape[1]):
        if c_orth[i,i] < 0:
            c_orth[:,i] *= -1
    return c_orth

del (ORTH_METHOD)
